""" PDB parsing class

    This module parses PDBs in accordance to PDB Format Description Version 2.2
    (1996); it is not very forgiving.   Each class in this module corresponds
    to a record in the PDB Format Description.  Much of the documentation for
    the classes is taken directly from the above PDB Format Description.

    ----------------------------

    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin;
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific
    Northwest National Laboratory, operated by Battelle Memorial Institute,
    Pacific Northwest Division for the U.S. Department Energy.;
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice,
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice,
		  this list of conditions and the following disclaimer in the documentation
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

__date__ = "4 August 2008"
__author__ = "Todd Dolinsky, Yong Huang"

import string, sys
import copy  ### PC
from errors import PDBInputError, PDBInternalError

lineParsers = {}

def RegisterLineParser(klass):
    lineParsers[klass.__name__] = klass
    return klass

class BaseRecord(object):
    """
    Base class for all records.
    Verifies the received record type
    """
    def __init__(self, line):
        record = string.strip(line[0:6])
        if record != self.__class__.__name__:
            raise ValueError, record

        self.original_text = line.rstrip('\r\n')

    def __str__(self):
        return self.original_text

@RegisterLineParser
class END(BaseRecord):
    """ END class

        The END records are paired with MODEL records to group individual
        structures found in a coordinate entry.
    """

    def __init__(self, line):
        """
            Initialize by parsing line (nothing to do)
        """
        super(END, self).__init__(line)

@RegisterLineParser
class MASTER(BaseRecord):
    """ MASTER class

        The MASTER record is a control record for bookkeeping. It lists the
        number of lines in the coordinate entry or file for selected record
        types.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD     DEFINITION
            -------------------------------------------------
            11-15    int    numRemark Number of REMARK records
            21-25    int    numHet    Number of HET records
            26-30    int    numHelix  Number of HELIX records
            31-35    int    numSheet  Number of SHEET records
            36-40    int    numTurn   Number of TURN records
            41-45    int    numSite   Number of SITE records
            46-50    int    numXform  Number of coordinate transformation
                                      records (ORIGX+SCALE+MTRIX)
            51-55    int    numCoord  Number of atomic coordinate records
                                      (ATOM+HETATM)
            56-60    int    numTer    Number of TER records
            61-65    int    numConect Number of CONECT records
            66-70    int    numSeq    Number of SEQRES records
        """
        super(MASTER, self).__init__(line)
        self.numRemark = int(string.strip(line[10:15]))
        self.numHet = int(string.strip(line[20:25]))
        self.numHelix = int(string.strip(line[25:30]))
        self.numSheet = int(string.strip(line[30:35]))
        self.numTurn = int(string.strip(line[35:40]))
        self.numSite = int(string.strip(line[40:45]))
        self.numXform = int(string.strip(line[45:50]))
        self.numCoord = int(string.strip(line[50:55]))
        self.numTer = int(string.strip(line[55:60]))
        self.numConect = int(string.strip(line[60:65]))
        self.numSeq = int(string.strip(line[65:70]))


@RegisterLineParser
class CONECT(BaseRecord):
    """ CONECT class

        The CONECT records specify connectivity between atoms for which
        coordinates are supplied. The connectivity is described using the atom
        serial number as found in the entry. CONECT records are mandatory for
        HET groups (excluding water) and for other bonds not specified in the
        standard residue connectivity table which involve atoms in standard
        residues (see Appendix 4 for the list of standard residues). These
        records are generated by the PDB.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD    DEFINITION
            --------------------------------------------
             7-11    int    serial   Atom serial number
            12-16    int    serial1  Serial number of bonded atom
            17-21    int    serial2  Serial number of bonded atom
            22-26    int    serial3  Serial number of bonded atom
            27-31    int    serial4  Serial number of bonded atom
            32-36    int    serial5  Serial number of hydrogen bonded atom
            37-41    int    serial6  Serial number of hydrogen bonded atom
            42-46    int    serial7  Serial number of salt bridged    atom
            47-51    int    serial8  Serial number of hydrogen bonded atom
            52-56    int    serial9  Serial number of hydrogen bonded atom
            57-61    int    serial10 Serial number of salt bridged    atom
        """
        super(CONECT, self).__init__(line)
        self.serial = int(string.strip(line[6:11]))
        try:  self.serial1 = int(string.strip(line[11:16]))
        except ValueError:  self.serial1 = None
        try:  self.serial2 = int(string.strip(line[16:21]))
        except ValueError:  self.serial2 = None
        try:  self.serial3 = int(string.strip(line[21:26]))
        except ValueError:  self.serial3 = None
        try:  self.serial4 = int(string.strip(line[26:31]))
        except ValueError:  self.serial4 = None
        try:  self.serial5 = int(string.strip(line[31:36]))
        except ValueError:  self.serial5 = None
        try:  self.serial6 = int(string.strip(line[36:41]))
        except ValueError:  self.serial6 = None
        try:  self.serial7 = int(string.strip(line[41:46]))
        except ValueError:  self.serial7 = None
        try:  self.serial8 = int(string.strip(line[46:51]))
        except ValueError:  self.serial8 = None
        try:  self.serial9 = int(string.strip(line[51:56]))
        except ValueError:  self.serial9 = None
        try:  self.serial10 = int(string.strip(line[56:61]))
        except ValueError:  self.serial10 = None

@RegisterLineParser
class NUMMDL(BaseRecord):
    """ NUMMDL class

        The NUMMDL record indicates total number of models in a PDB entry.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE      FIELD         DEFINITION
            -----------------------------------------------------------
            11-14    int       modelNumber   Number of models.
        """
        super(NUMMDL, self).__init__(line)
        try:
            self.modelNumber = int(string.strip(line[10:14]))
        except ValueError:
            self.modelNumber = None

@RegisterLineParser
class ENDMDL(BaseRecord):
    """ ENDMDL class

        The ENDMDL records are paired with MODEL records to group individual
        structures found in a coordinate entry.
    """

    def __init__(self, line):
        """
            Initialize by parsing line (nothing to do)
        """
        super(ENDMDL, self).__init__(line)

@RegisterLineParser
class TER(BaseRecord):
    """ TER class

        The TER record indicates the end of a list of ATOM/HETATM records for a
        chain.
    """

    def __init__(self, line):
        """ Initialize by parsing line:

            COLUMNS  TYPE   FIELD   DEFINITION
            -------------------------------------------
             7-11    int    serial  Serial number.
            18-20    string resName Residue name.
            22       string chainID Chain identifier.
            23-26    int    resSeq  Residue sequence number.
            27       string iCode   Insertion code.
        """
        super(TER, self).__init__(line)
        try: # Not really needed
            self.serial = int(string.strip(line[6:11]))
            self.resName = string.strip(line[17:20])
            self.chainID = string.strip(line[21])
            self.resSeq = int(string.strip(line[22:26]))
            self.iCode = string.strip(line[26])
        except (IndexError, ValueError):
            self.serial = None
            self.resName = None
            self.chainID = None
            self.resSeq = None
            self.iCode = None

@RegisterLineParser
class SIGUIJ(BaseRecord):
    """ SIGUIJ class

        The SIGUIJ records present the anisotropic temperature factors.
    """

    def __init__(self, line):
        """
            Initialize by parsing line:

              COLUMNS  TYPE   FIELD   DEFINITION
              ------------------------------------------------------
               7-11    int    serial  Atom serial number.
              13-16    string name    Atom name.
              17       string altLoc  Alternate location indicator.
              18-20    string resName Residue name.
              22       string chainID Chain identifier.
              23-26    int    resSeq  Residue sequence number.
              27       string iCode   Insertion code.
              29-35    int    sig11   Sigma U(1,1)
              36-42    int    sig22   Sigma U(2,2)
              43-49    int    sig33   Sigma U(3,3)
              50-56    int    sig12   Sigma U(1,2)
              57-63    int    sig13   Sigma U(1,3)
              64-70    int    sig23   Sigma U(2,3)
              73-76    string segID   Segment identifier, left-justified.
              77-78    string element Element symbol, right-justified.
              79-80    string charge  Charge on the atom.
        """
        super(SIGUIJ, self).__init__(line)

        self.serial = int(string.strip(line[6:11]))
        self.name = string.strip(line[12:16])
        self.altLoc = string.strip(line[16])
        self.resName = string.strip(line[17:20])
        self.chainID = string.strip(line[21])
        self.resSeq = int(string.strip(line[22:26]))
        self.iCode = string.strip(line[26])
        self.sig11 = int(string.strip(line[28:35]))
        self.sig22 = int(string.strip(line[35:42]))
        self.sig33 = int(string.strip(line[42:49]))
        self.sig12 = int(string.strip(line[49:56]))
        self.sig13 = int(string.strip(line[56:63]))
        self.sig23 = int(string.strip(line[63:70]))
        self.segID = string.strip(line[72:76])
        self.element = string.strip(line[76:78])
        self.charge = string.strip(line[78:80])


@RegisterLineParser
class ANISOU(BaseRecord):
    """ ANISOU class

        The ANISOU records present the anisotropic temperature factors.
    """

    def __init__(self, line):
        """
            Initialize by parsing line:

              COLUMNS  TYPE   FIELD   DEFINITION
              ------------------------------------------------------
               7-11    int    serial  Atom serial number.
              13-16    string name    Atom name.
              17       string altLoc  Alternate location indicator.
              18-20    string resName Residue name.
              22       string chainID Chain identifier.
              23-26    int    resSeq  Residue sequence number.
              27       string iCode   Insertion code.
              29-35    int    u00     U(1,1)
              36-42    int    u11     U(2,2)
              43-49    int    u22     U(3,3)
              50-56    int    u01     U(1,2)
              57-63    int    u02     U(1,3)
              64-70    int    u12     U(2,3)
              73-76    string segID   Segment identifier, left-justified.
              77-78    string element Element symbol, right-justified.
              79-80    string charge  Charge on the atom.
        """
        super(ANISOU, self).__init__(line)
        self.serial = int(string.strip(line[6:11]))
        self.name = string.strip(line[12:16])
        self.altLoc = string.strip(line[16])
        self.resName = string.strip(line[17:20])
        self.chainID = string.strip(line[21])
        self.resSeq = int(string.strip(line[22:26]))
        self.iCode = string.strip(line[26])
        self.u00 = int(string.strip(line[28:35]))
        self.u11 = int(string.strip(line[35:42]))
        self.u22 = int(string.strip(line[42:49]))
        self.u01 = int(string.strip(line[49:56]))
        self.u02 = int(string.strip(line[56:63]))
        self.u12 = int(string.strip(line[63:70]))
        self.segID = string.strip(line[72:76])
        self.element = string.strip(line[76:78])
        self.charge = string.strip(line[78:80])

@RegisterLineParser
class SIGATM(BaseRecord):
    """ SIGATM class

        The SIGATM records present the standard deviation of atomic parameters
        as they appear in ATOM and HETATM records.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD    DEFINITION
            ---------------------------------------------
            7-11      int   serial   Atom serial number.
            13-16     string name    Atom name.
            17        string altLoc  Alternate location indicator.
            18-20     string resName Residue name.
            22        string chainID Chain identifier.
            23-26     int    resSeq  Residue sequence number.
            27        string iCode   Code for insertion of residues.
            31-38     float  sigX    Standard devition of orthogonal
                                     coordinates for X in Angstroms.
            39-46     float  sigY    Standard devition of orthogonal
                                     coordinates for Y in Angstroms.
            47-54     float  sigZ    Standard devition of orthogonal
                                     coordinates for Z in Angstroms.
            55-60     float  sigOcc  Standard devition of occupancy.
            61-66     float  sigTemp Standard devition of temperature factor.
            73-76     string segID   Segment identifier, left-justified.
            77-78     string element Element symbol, right-justified.
            79-80     string charge  Charge on the atom.
        """
        super(SIGATM, self).__init__(line)
        self.serial = int(string.strip(line[6:11]))
        self.name = string.strip(line[12:16])
        self.altLoc = string.strip(line[16])
        self.resName = string.strip(line[17:20])
        self.chainID = string.strip(line[21])
        self.resSeq = int(string.strip(line[22:26]))
        self.iCode = string.strip(line[26])
        self.sigX = float(string.strip(line[30:38]))
        self.sigY = float(string.strip(line[38:46]))
        self.sigZ = float(string.strip(line[46:54]))
        self.sigOcc = float(string.strip(line[54:60]))
        self.sigTemp = float(string.strip(line[60:66]))
        self.segID = string.strip(line[72:76])
        self.element = string.strip(line[76:78])
        self.charge = string.strip(line[78:80])

@RegisterLineParser
class HETATM(BaseRecord):
    """ HETATM class

        The HETATM records present the atomic coordinate records for atoms
        within "non-standard" groups. These records are used for water
        molecules and atoms presented in HET groups.
    """

    def __init__(self,line,sybylType="A.aaa",lBonds=[],lBondedAtoms=[]): ### PC
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------------------
            7-11      int   serial        Atom serial number.
            13-16     string name          Atom name.
            17        string altLoc        Alternate location indicator.
            18-20     string resName       Residue name.
            22        string chainID       Chain identifier.
            23-26     int    resSeq        Residue sequence number.
            27        string iCode         Code for insertion of residues.
            31-38     float  x             Orthogonal coordinates for X in
                                           Angstroms.
            39-46     float  y             Orthogonal coordinates for Y in
                                           Angstroms.
            47-54     float  z             Orthogonal coordinates for Z in
                                           Angstroms.
            55-60     float  occupancy     Occupancy.
            61-66     float  tempFactor    Temperature factor.
            73-76     string segID         Segment identifier, left-justified.
            77-78     string element       Element symbol, right-justified.
            79-80     string charge        Charge on the atom.
        """
        super(HETATM, self).__init__(line)
        self.serial = int(string.strip(line[6:11]))
        self.name = string.strip(line[12:16])
        self.altLoc = string.strip(line[16])
        try:
            self.resName = string.strip(line[17:20])
            self.chainID = string.strip(line[21])
            self.resSeq = int(string.strip(line[22:26]))
            self.iCode = string.strip(line[26])
        except:
            raise ValueError, 'Residue name must be less than 4 characters!'
        self.x = float(string.strip(line[30:38]))
        self.y = float(string.strip(line[38:46]))
        self.z = float(string.strip(line[46:54]))
        ### PC
#            self.lAtoms = lAtoms
        self.sybylType = sybylType
        self.lBondedAtoms = lBondedAtoms
        self.lBonds = lBonds
        self.radius = 1.0
        self.isCterm=0
        self.isNterm=0
        ###
        try:
            self.occupancy = float(string.strip(line[54:60]))
            self.tempFactor = float(string.strip(line[60:66]))
            self.segID = string.strip(line[72:76])
            self.element = string.strip(line[76:78])
            self.charge = string.strip(line[78:80])
        except (ValueError, IndexError):
            self.occupancy = 0.00
            self.tempFactor = 0.00
            self.segID = ""
            self.element = ""
            self.charge = ""

#     def __str__(self):
#         """
#             Print object as string
#
#             COLUMNS  TYPE   FIELD  DEFINITION
#             ---------------------------------------------
#             7-11      int   serial        Atom serial number.
#             13-16     string name          Atom name.
#             17        string altLoc        Alternate location indicator.
#             18-20     string resName       Residue name.
#             22        string chainID       Chain identifier.
#             23-26     int    resSeq        Residue sequence number.
#             27        string iCode         Code for insertion of residues.
#             31-38     float  x             Orthogonal coordinates for X in
#                                            Angstroms.
#             39-46     float  y             Orthogonal coordinates for Y in
#                                            Angstroms.
#             47-54     float  z             Orthogonal coordinates for Z in
#                                            Angstroms.
#             55-60     float  occupancy     Occupancy.
#             61-66     float  tempFactor    Temperature factor.
#             73-76     string segID         Segment identifier, left-justified.
#             77-78     string element       Element symbol, right-justified.
#             79-80     string charge        Charge on the atom.
#         """
#         str = ""
#         tstr = "HETATM"
#         str = str + string.ljust(tstr, 6)[:6]
#         tstr = "%d" % self.serial
#         str = str + string.rjust(tstr, 5)[:5]
#         str = str + " "
#         tstr = self.name
#         if len(tstr) == 4:
#             str = str + string.ljust(tstr, 4)[:4]
#         else:
#             str = str + " " + string.ljust(tstr, 3)[:3]
#         tstr = self.altLoc
#         str = str + string.ljust(tstr, 1)[:1]
#         tstr = self.resName
#         str = str + string.ljust(tstr, 3)[:3]
#         str = str + " "
#         tstr = self.chainID
#         str = str + string.ljust(tstr, 1)[:1]
#         tstr = "%d" % self.resSeq
#         str = str + string.rjust(tstr, 4)[:4]
#         tstr = self.iCode
#         str = str + string.ljust(tstr, 1)[:1]
#         str = str + "   "
#         tstr = "%8.3f" % self.x
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%8.3f" % self.y
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%8.3f" % self.z
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%6.2f" % self.occupancy
#         str = str + string.ljust(tstr, 6)[:6]
#         tstr = "%6.2f" % self.tempFactor
#         str = str + string.rjust(tstr, 6)[:6]
#         tstr = self.segID
#         str = str + string.ljust(tstr, 4)[:4]
#         tstr = self.element
#         str = str + string.ljust(tstr, 2)[:2]
#         tstr = self.charge
#         str = str + string.ljust(tstr, 2)[:2]
#         return str

### PC
# to do:  - parse SUBSTRUCTURE
#         - avoid/detect blanks in @<TRIPOS>BOND
#         - what happens, if no SUBSTRUCTURE present?
#         - different order of SUBSTRUCTURE/MOLECULE
#         - readlines instead of read -> blanks are avoided (you get a list)
#         - (maybe) flag for parsing each RTI

class MOL2BOND:
    """
    Bonding of MOL2 files
    """
    def __init__(self, frm, to, type, id=0):
        self.to   = to     # bond to this atom
        self.frm  = frm    # bond from atom
        self.type = type   # 1=single, 2=double, ar=aromatic
        self.id   = id     # bond_id

class MOL2MOLECULE:
    """
    Tripos MOL2 molecule
    For further information look at (web page exists: 25 August 2005):
    http://www.tripos.com/index.php?family=modules,SimplePage,,,&page=sup_mol2&s=0
    """
    def __init__(self):
        self.lAtoms         = []       # all atoms of class <ATOM>
        self.lBonds         = []       # all bonds of class <BOND>
        self.lPDBAtoms      = []       # PDB-like list of all atoms

    def read(self,file):
        """
        Routines for reading MOL2 file
        """
        #self.filename = filename
        #data = open(self.filename).read()

        data = file.read()
        data = data.replace("\r\n", "\n")
        data = data.replace("\r", "\n")

        # ATOM section
        start = data.find("@<TRIPOS>ATOM")
        stop = data.find("@<TRIPOS>BOND")

        # Do some error checking
        if start == -1:
            raise PDBInputError("Unable to find '@<TRIPOS>ATOM' in MOL2 file!")
        elif stop == -1:
            raise PDBInputError("Unable to find '@<TRIPOS>BOND' in MOL2 file!")

        atoms = data[start+14:stop-2].split("\n")
        # BOND section
        start = data.find("@<TRIPOS>BOND")
        stop = data.find("@<TRIPOS>SUBSTRUCTURE")

        # More error checking
        if stop == -1:
            raise PDBInputError("Unable to find '@<TRIPOS>SUBSTRUCTURE' in MOL2 file!")

        bonds = data[start+14:stop-1].split("\n")
        self.parseAtoms(atoms)
        self.parseBonds(bonds)
        self.createlBondedAtoms()
        #self.createPDBlineFromMOL2(atoms)

    def parseAtoms(self,AtomList):
        """
        for parsing @<TRIPOS>ATOM
        """
        for AtomLine in AtomList:
            SeparatedAtomLine = AtomLine.split()
            # Special handling for blank lines
            if len(SeparatedAtomLine) == 0:
                continue

            # Error checking
            if len(SeparatedAtomLine) < 8:
                raise PDBInputError("Bad atom entry in MOL2 file: %s" % AtomLine)

            fakeRecord = "HETATM"
            fakeChain = " L"
            try:
                mol2pdb = '%s%5i%5s%4s%2s%4i    %8.3f%8.3f%8.3f' %\
                   (fakeRecord,int(SeparatedAtomLine[0]),
                    SeparatedAtomLine[1],SeparatedAtomLine[7][:4],
                    fakeChain,int(SeparatedAtomLine[6]),
                    float(SeparatedAtomLine[2]),float(SeparatedAtomLine[3]),
                    float(SeparatedAtomLine[4]))
            except ValueError:
                raise PDBInputError("Bad atom entry in MOL2 file: %s" % AtomLine)
            thisAtom = HETATM(mol2pdb, SeparatedAtomLine[5],[],[])
            if len(SeparatedAtomLine)>8:
                charge=SeparatedAtomLine[8]
                try:
                    thisAtom.mol2charge=float(charge)
                except:
                    print 'Warning. Non-float charge in mol2 file.',charge
                    thisAtom.mol2charge=None
            self.lPDBAtoms.append(mol2pdb)
            self.lAtoms.append(thisAtom)

    def parseBonds(self,BondList):
        """
        for parsing @<TRIPOS>BOND
        """
        for BondLine in BondList:
            SeparatedBondLine = BondLine.split()
            # Special handling for blank lines
            if len(SeparatedBondLine) == 0:
                continue
            if len(SeparatedBondLine) < 4:
                raise PDBInputError("Bad bond entry in MOL2 file: %s" % BondLine)
            try:
                thisBond = MOL2BOND(
                    int(SeparatedBondLine[1]), # bond frm
                    int(SeparatedBondLine[2]), # bond to
                    SeparatedBondLine[3],      # bond type
                    int(SeparatedBondLine[0])  # bond id
                    )
            except ValueError:
                raise PDBInputError("Bad bond entry in MOL2 file: %s" % BondLine)
            self.lBonds.append(thisBond)

    def createlBondedAtoms(self):
        """
        Creates for each atom a list of the bonded Atoms

        This becomes one attribute of MOL2ATOM!
        """
        for bond in self.lBonds:
            self.lAtoms[bond.frm-1].lBondedAtoms.append(
                self.lAtoms[bond.to-1])

            self.lAtoms[bond.to-1].lBondedAtoms.append(
                self.lAtoms[bond.frm-1])

            atbond = copy.deepcopy(bond)
            atbond.other_atom=self.lAtoms[bond.to-1]
            self.lAtoms[bond.frm-1].lBonds.append(atbond)

            atbond = copy.deepcopy(bond)
            atbond.other_atom=self.lAtoms[bond.frm-1]
            self.lAtoms[bond.to-1].lBonds.append(atbond)
        return


    def createPDBlineFromMOL2(self):
        FakeType = "HETATM"
        return ('%s%5i%5s%4s%2s%5s   %8.3f%8.3f%8.3f\n' %
                (FakeType,            self.serial,   self.name,
                 self.resName, ' L',          self.resSeq,
                 self.x,self.y, self.z))
### PC

@RegisterLineParser
class ATOM(BaseRecord):
    """ ATOM class

        The ATOM records present the atomic coordinates for standard residues.
        They also present the occupancy and temperature factor for each atom.
        Heterogen coordinates use the HETATM record type. The element symbol is
        always present on each ATOM record; segment identifier and charge are
        optional.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------------------
            7-11      int   serial        Atom serial number.
            13-16     string name          Atom name.
            17        string altLoc        Alternate location indicator.
            18-20     string resName       Residue name.
            22        string chainID       Chain identifier.
            23-26     int    resSeq        Residue sequence number.
            27        string iCode         Code for insertion of residues.
            31-38     float  x             Orthogonal coordinates for X in
                                           Angstroms.
            39-46     float  y             Orthogonal coordinates for Y in
                                           Angstroms.
            47-54     float  z             Orthogonal coordinates for Z in
                                           Angstroms.
            55-60     float  occupancy     Occupancy.
            61-66     float  tempFactor    Temperature factor.
            73-76     string segID         Segment identifier, left-justified.
            77-78     string element       Element symbol, right-justified.
            79-80     string charge        Charge on the atom.
        """
        super(ATOM, self).__init__(line)
        self.serial = int(string.strip(line[6:11]))
        self.name = string.strip(line[12:16])
        self.altLoc = string.strip(line[16])
        self.resName = string.strip(line[17:20])
        self.chainID = string.strip(line[21])
        self.resSeq = int(string.strip(line[22:26]))
        self.iCode = string.strip(line[26])
        self.x = float(string.strip(line[30:38]))
        self.y = float(string.strip(line[38:46]))
        self.z = float(string.strip(line[46:54]))
        try:
            self.occupancy = float(string.strip(line[54:60]))
            self.tempFactor = float(string.strip(line[60:66]))
            self.segID = string.strip(line[72:76])
            self.element = string.strip(line[76:78])
            self.charge = string.strip(line[78:80])
        except (ValueError, IndexError):
            self.occupancy = 0.00
            self.tempFactor = 0.00
            self.segID = ""
            self.element = ""
            self.charge = ""

#     def __str__(self):
#         """
#             Print object as string
#
#             COLUMNS  TYPE   FIELD  DEFINITION
#             ---------------------------------------------
#             7-11      int   serial        Atom serial number.
#             13-16     string name          Atom name.
#             17        string altLoc        Alternate location indicator.
#             18-20     string resName       Residue name.
#             22        string chainID       Chain identifier.
#             23-26     int    resSeq        Residue sequence number.
#             27        string iCode         Code for insertion of residues.
#             31-38     float  x             Orthogonal coordinates for X in
#                                            Angstroms.
#             39-46     float  y             Orthogonal coordinates for Y in
#                                            Angstroms.
#             47-54     float  z             Orthogonal coordinates for Z in
#                                            Angstroms.
#             55-60     float  occupancy     Occupancy.
#             61-66     float  tempFactor    Temperature factor.
#             73-76     string segID         Segment identifier, left-justified.
#             77-78     string element       Element symbol, right-justified.
#             79-80     string charge        Charge on the atom.
#         """
#         str = ""
#         tstr = "ATOM"
#         str = str + string.ljust(tstr, 6)[:6]
#         tstr = "%d" % self.serial
#         str = str + string.rjust(tstr, 5)[:5]
#         str = str + " "
#         tstr = self.name
#         if len(tstr) == 4:
#             str = str + string.ljust(tstr, 4)[:4]
#         else:
#             str = str + " " + string.ljust(tstr, 3)[:3]
#         tstr = self.altLoc
#         str = str + string.ljust(tstr, 1)[:1]
#         tstr = self.resName
#         str = str + string.ljust(tstr, 3)[:3]
#         str = str + " "
#         tstr = self.chainID
#         str = str + string.ljust(tstr, 1)[:1]
#         tstr = "%d" % self.resSeq
#         str = str + string.rjust(tstr, 4)[:4]
#         tstr = self.iCode
#         str = str + string.ljust(tstr, 1)[:1]
#         str = str + "   "
#         tstr = "%8.3f" % self.x
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%8.3f" % self.y
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%8.3f" % self.z
#         str = str + string.ljust(tstr, 8)[:8]
#         tstr = "%6.2f" % self.occupancy
#         str = str + string.ljust(tstr, 6)[:6]
#         tstr = "%6.2f" % self.tempFactor
#         str = str + string.ljust(tstr, 6)[:6]
#         tstr = self.segID
#         str = str + string.ljust(tstr, 4)[:4]
#         tstr = self.element
#         str = str + string.ljust(tstr, 2)[:2]
#         tstr = self.charge
#         str = str + string.ljust(tstr, 2)[:2]
#         return str

@RegisterLineParser
class MODEL(BaseRecord):
    """ MODEL class

        The MODEL record specifies the model serial number when multiple
        structures are presented in a single coordinate entry, as is often the
        case with structures determined by NMR.
    """

    def __init__(self, line):
        """
           Initialize by parsing line

           COLUMNS  TYPE   FIELD  DEFINITION
           -----------------------------------------------------
           11-14    int    serial Model serial number.
        """
        super(MODEL, self).__init__(line)
        self.serial = int(string.strip(line[10:14]))

@RegisterLineParser
class TVECT(BaseRecord):
    """ TVECT class

        The TVECT records present the translation vector for infinite
        covalently connected structures.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------
             8-10    int    serial Serial number
            11-20    float  t1     Components of translation vector
            21-30    float  t2     Components of translation vector
            31-40    float  t2     Components of translation vector
            41-70    string text   Comments
        """
        super(TVECT, self).__init__(line)
        self.serial = int(string.strip(line[7:10]))
        self.t1 = float(string.strip(line[10:20]))
        self.t2 = float(string.strip(line[20:30]))
        self.t3 = float(string.strip(line[30:40]))
        self.text = string.strip(line[40:70])


class MTRIXn(BaseRecord):
    """ MTRIXn baseclass

        The MTRIXn (n = 1, 2, or 3) records present transformations expressing
        non-crystallographic symmetry.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------
             8-10    int    serial Serial number
            11-20    float  mn1    M31
            21-30    float  mn2    M32
            31-40    float  mn3    M33
            46-55    float  vn     V3
            60       int    iGiven 1 if coordinates for the representations
                            which are approximately related by the
                            transformations of the molecule are contained in
                            the entry.  Otherwise, blank.
        """
        super(MTRIXn, self).__init__(line)
        self.serial = int(string.strip(line[7:10]))
        self.mn1 = float(string.strip(line[10:20]))
        self.mn2 = float(string.strip(line[20:30]))
        self.mn3 = float(string.strip(line[30:40]))
        self.vn = float(string.strip(line[45:55]))
        try:
            self.iGiven = int(string.strip(line[59]))
        except (ValueError, IndexError):
            self.iGiven = None

@RegisterLineParser
class MTRIX3(MTRIXn):
    pass

@RegisterLineParser
class MTRIX2(MTRIXn):
    pass

@RegisterLineParser
class MTRIX1(MTRIXn):
    pass

class SCALEn(BaseRecord):
    """ SCALEn baseclass

        The SCALEn (n = 1, 2, or 3) records present the transformation from the
        orthogonal coordinates as contained in the entry to fractional
        crystallographic coordinates. Non-standard coordinate systems should be
        explained in the remarks.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------
            11-20    float  sn1    S31
            21-30    float  sn2    S32
            31-40    float  sn3    S33
            46-55    float  un     U3
        """
        super(SCALEn, self).__init__(line)
        self.sn1 = float(string.strip(line[10:20]))
        self.sn2 = float(string.strip(line[20:30]))
        self.sn3 = float(string.strip(line[30:40]))
        self.un = float(string.strip(line[45:55]))

@RegisterLineParser
class SCALE3(SCALEn):
    pass

@RegisterLineParser
class SCALE2(SCALEn):
    pass

@RegisterLineParser
class SCALE1(SCALEn):
    pass

class ORIGXn(BaseRecord):
    """ ORIGX2 class

        The ORIGXn (n = 1, 2, or 3) records present the transformation from the
        orthogonal coordinates contained in the entry to the submitted
        coordinates.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------
            11-20    float  on1    O21
            21-30    float  on2    O22
            31-40    float  on3    O23
            46-55    float  tn     T2
        """
        super(ORIGXn, self).__init__(line)
        self.on1 = float(string.strip(line[10:20]))
        self.on2 = float(string.strip(line[20:30]))
        self.on3 = float(string.strip(line[30:40]))
        self.tn = float(string.strip(line[45:55]))

@RegisterLineParser
class ORIGX2(ORIGXn):
    pass

@RegisterLineParser
class ORIGX3(ORIGXn):
    pass

@RegisterLineParser
class ORIGX1(ORIGXn):
    pass

@RegisterLineParser
class CRYST1(BaseRecord):
    """ CRYST1 class

        The CRYST1 record presents the unit cell parameters, space group, and Z
        value. If the structure was not determined by crystallographic means,
        CRYST1 simply defines a unit cube.
    """

    def __init__(self, line):
        """
           Initialize by parsing line

           COLUMNS  TYPE   FIELD  DEFINITION
           ---------------------------------------
            7-15    float  a      a (Angstroms).
           16-24    float  b      b (Angstroms).
           25-33    float  c      c (Angstroms).
           34-40    float  alpha  alpha (degrees).
           41-47    float  beta   beta (degrees).
           48-54    float  gamma  gamma (degrees).
           56-66    string sGroup Space group.
           67-70    int    z      Z value.
        """
        super(CRYST1, self).__init__(line)
        self.a = float(string.strip(line[6:15]))
        self.b = float(string.strip(line[15:24]))
        self.c = float(string.strip(line[24:33]))
        self.alpha = float(string.strip(line[33:40]))
        self.beta = float(string.strip(line[40:47]))
        self.gamma = float(string.strip(line[47:54]))
        self.sGroup = string.strip(line[55:65])
        self.z = int(string.strip(line[66:70]))


@RegisterLineParser
class SITE(BaseRecord):
    """ SITE class

        The SITE records supply the identification of groups comprising
        important sites in the macromolecule.
    """

    def __init__(self, line):
        """
            Initialize by parsing the line

            COLUMNS  TYPE   FIELD    DEFINITION
            --------------------------------------------------------------
             8-10    int    seqNum   Sequence number.
            12-14    string siteID   Site name.
            16-17    int    numRes   Number of residues comprising site.
            19-21    string resName1 Residue name for first residue
                                     comprising site.
            23       string chainID1 Chain identifier for first residue
                                     comprising site.
            24-27    int    seq1     Residue sequence number for first
                                     residue comprising site.
            28       string iCode1   Insertion code for first residue
                                     comprising site.
            30-32    string resName2 Residue name for second residue
                                     comprising site.
            34       string chainID2 Chain identifier for second residue
                                     comprising site.
            35-38    int    seq2     Residue sequence number for second
                                     residue comprising site.
            39       string iCode2   Insertion code for second residue
                                     comprising site.
            41-43    string resName3 Residue name for third residue
                                     comprising site.
            45       string chainID3 Chain identifier for third residue
                                     comprising site.
            46-49    int    seq3     Residue sequence number for third
                                     residue comprising site.
            50       string iCode3   Insertion code for third residue
                                     comprising site.
            52-54    string resName4 Residue name for fourth residue
                                     comprising site.
            56       string chainID4 Chain identifier for fourth residue
                                     comprising site.
            57-60    int    seq4     Residue sequence number for fourth
                                     residue comprising site.
            61       string iCode4   Insertion code for fourth residue
                                     comprising site.
        """
        super(SITE, self).__init__(line)
        self.seqNum = int(string.strip(line[7:10]))
        self.siteID = string.strip(line[11:14])
        self.numRes = int(string.strip(line[15:17]))
        self.resName1 = string.strip(line[18:21])
        self.chainID1 = string.strip(line[22])
        self.seq1 = int(string.strip(line[23:27]))
        self.iCode1 = string.strip(line[27])
        self.resName2 = string.strip(line[29:32])
        self.chainID2 = string.strip(line[33])
        self.seq2 = int(string.strip(line[34:38]))
        self.iCode2 = string.strip(line[38])
        self.resName3 = string.strip(line[40:43])
        self.chainID3 = string.strip(line[44])
        self.seq3 = int(string.strip(line[45:49]))
        self.iCode3 = string.strip(line[49])
        self.resName4 = string.strip(line[51:54])
        self.chainID4 = string.strip(line[55])
        self.seq4 = int(string.strip(line[56:60]))
        try:
            self.iCode4 = string.strip(line[60])
        except IndexError:
            self.iCode4 = None

@RegisterLineParser
class CISPEP(BaseRecord):
    """ CISPEP field

        CISPEP records specify the prolines and other peptides found to be in
        the cis conformation. This record replaces the use of footnote records
        to list cis peptides.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD    DEFINITION
            -----------------------------------------------------------
            8-10     int    serNum   Record serial number.
            12-14    string pep1     Residue name.
            16       string chainID1 Chain identifier.
            18-21    int    seqNum1  Residue sequence number.
            22       string icode1   Insertion code.
            26-28    string pep2     Residue name.
            30       string chainID2 Chain identifier.
            32-35    int    seqNum2  Residue sequence number.
            36       string icode2   Insertion code.
            44-46    int    modNum   Identifies the specific model.
            54-59    float  measure  Measure of the angle in degrees.
        """
        super(CISPEP, self).__init__(line)
        self.serNum = int(string.strip(line[7:10]))
        self.pep1 = string.strip(line[11:14])
        self.chainID1 = string.strip(line[15])
        self.seqNum1 = int(string.strip(line[17:21]))
        self.icode1 = string.strip(line[21])
        self.pep2 = string.strip(line[25:28])
        self.chainID2 = string.strip(line[29])
        self.seqNum2 = int(string.strip(line[31:35]))
        self.icode2 = string.strip(line[35])
        self.modNum = int(string.strip(line[43:46]))
        self.measure = float(string.strip(line[53:59]))

@RegisterLineParser
class SLTBRG(BaseRecord):
    """ SLTBRG field

        The SLTBRG records specify salt bridges in the entry.
        records and is provided here for convenience in searching.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD     DEFINITION
            -----------------------------------------------------
            13-16    string name1     Atom name.
            17       string altLoc1   Alternate location indicator.
            18-20    string resName1  Residue name.
            22       string chainID1  Chain identifier.
            23-26    int    resSeq1   Residue sequence number.
            27       string iCode1    Insertion code.
            43-46    string name2     Atom name.
            47       string altLoc2   Alternate location indicator.
            48-50    string resName2  Residue name.
            52       string chainID2  Chain identifier.
            53-56    int    resSeq2   Residue sequence number.
            57       string iCode2    Insertion code.
            60-65    string sym1      Symmetry operator for 1st atom.
            67-72    string sym2      Symmetry operator for 2nd atom.
        """
        super(SLTBRG, self).__init__(line)
        self.name1 = string.strip(line[12:16])
        self.altLoc1 = string.strip(line[16])
        self.resName1 = string.strip(line[17:20])
        self.chainID1 = string.strip(line[21])
        self.resSeq1 = int(string.strip(line[22:26]))
        self.iCode1 = string.strip(line[26])
        self.name2 = string.strip(line[42:46])
        self.altLoc2 = string.strip(line[46])
        self.resName2 = string.strip(line[47:50])
        self.chainID2 = string.strip(line[51])
        self.resSeq2 = int(string.strip(line[52:56]))
        self.iCode2 = string.strip(line[56])
        self.sym1 = string.strip(line[59:65])
        self.sym2 = string.strip(line[66:72])

@RegisterLineParser
class HYDBND(BaseRecord):
    """ HYDBND field

        The HYDBND records specify hydrogen bonds in the entry.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            -----------------------------------------------------------
            13-16    string name1          Atom name.
            17       string altLoc1        Alternate location indicator.
            18-20    string resName1       Residue name.
            22       string Chain1         Chain identifier.
            23-27    int    resSeq1        Residue sequence number.
            28       string ICode1         Insertion code.
            30-33    string nameH          Hydrogen atom name.
            34       string altLocH        Alternate location indicator.
            36       string ChainH         Chain identifier.
            37-41    int    resSeqH        Residue sequence number.
            42       string iCodeH         Insertion code.
            44-47    string name2          Atom name.
            48       string altLoc2        Alternate location indicator.
            49-51    string resName2       Residue name.
            53       string chainID2       Chain identifier.
            54-58    int    resSeq2        Residue sequence number.
            59       string iCode2         Insertion code.
            60-65    string sym1           Symmetry operator for 1st
                                                          non-hydrogen atom.
            67-72    string sym2           Symmetry operator for 2nd
                                              non-hydrogen atom.
        """
        super(HYDBND, self).__init__(line)
        self.name1 = string.strip(line[12:16])
        self.altLoc1 = string.strip(line[16])
        self.resName1 = string.strip(line[17:20])
        self.Chain1 = string.strip(line[21])
        self.resSeq1 = string.strip(line[22:27])
        self.ICode1 = string.strip(line[27])
        self.nameH = string.strip(line[29:33])
        self.altLocH = string.strip(line[33])
        self.ChainH = string.strip(line[35])
        self.resSeqH = string.strip(line[36:41])
        self.ICodeH = string.strip(line[41])
        self.name2 = string.strip(line[43:47])
        self.altLoc2 = string.strip(line[47])
        self.resName2 = string.strip(line[48:51])
        self.Chain2 = string.strip(line[52])
        self.resSeq2 = string.strip(line[53:58])
        self.ICode2 = string.strip(line[58])
        self.sym1 = string.strip(line[59:65])
        self.sym2 = string.strip(line[66:72])

@RegisterLineParser
class LINK(BaseRecord):
    """ LINK field

        The LINK records specify connectivity between residues that is not
        implied by the primary structure. Connectivity is expressed in terms of
        the atom names. This record supplements information given in CONECT
        records and is provided here for convenience in searching.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD     DEFINITION
            -----------------------------------------------------
            13-16    string name1     Atom name.
            17       string altLoc1   Alternate location indicator.
            18-20    string resName1  Residue name.
            22       string chainID1  Chain identifier.
            23-26    int    resSeq1   Residue sequence number.
            27       string iCode1    Insertion code.
            43-46    string name2     Atom name.
            47       string altLoc2   Alternate location indicator.
            48-50    string resName2  Residue name.
            52       string chainID2  Chain identifier.
            53-56    int    resSeq2   Residue sequence number.
            57       string iCode2    Insertion code.
            60-65    string sym1      Symmetry operator for 1st atom.
            67-72    string sym2      Symmetry operator for 2nd atom.
        """
        super(LINK, self).__init__(line)
        self.name1 = string.strip(line[12:16])
        self.altLoc1 = string.strip(line[16])
        self.resName1 = string.strip(line[17:20])
        self.chainID1 = string.strip(line[21])
        self.resSeq1 = int(string.strip(line[22:26]))
        self.iCode1 = string.strip(line[26])
        self.name2 = string.strip(line[42:46])
        self.altLoc2 = string.strip(line[46])
        self.resName2 = string.strip(line[47:50])
        self.chainID2 = string.strip(line[51])
        self.resSeq2 = int(string.strip(line[52:56]))
        self.iCode2 = string.strip(line[56])
        self.sym1 = string.strip(line[59:65])
        self.sym2 = string.strip(line[66:72])


@RegisterLineParser
class SSBOND(BaseRecord):
    """ SSBOND field

        The SSBOND record identifies each disulfide bond in protein and
        polypeptide structures by identifying the two residues involved in the
        bond.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            -----------------------------------------------------
             8 - 10  int    serNum         Serial number.
            16       string chainID1       Chain identifier.
            18 - 21  int    seqNum1        Residue sequence number.
            22       string icode1         Insertion code.
            30       string chainID2       Chain identifier.
            32 - 35  int    seqNum2        Residue sequence number.
            36       string icode2         Insertion code.
            60 - 65  string sym1           Symmetry operator for 1st residue.
            67 - 72  string sym2           Symmetry operator for 2nd residue.
        """
        super(SSBOND, self).__init__(line)
        self.serNum = int(string.strip(line[7:10]))
        self.chainID1 = string.strip(line[15])
        self.seqNum1 = int(string.strip(line[17:21]))
        self.icode1 = string.strip(line[21])
        self.chainID2 = string.strip(line[29])
        self.seqNum2 = int(string.strip(line[31:35]))
        self.icode2 = string.strip(line[35])
        self.sym1 = string.strip(line[59:65])
        self.sym2 = string.strip(line[66:72])

@RegisterLineParser
class TURN(BaseRecord):
    """ TURN field

        The TURN records identify turns and other short loop turns which
        normally connect other secondary structure segments.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD       DEFINITION
            ---------------------------------------------------------
            8-10     int    seq         Turn number; starts with 1 and
                                        increments by one.
            12-14    string turnId      Turn identifier
            16-18    string initResName Residue name of initial residue in
                                        turn.
            20       string initChainId Chain identifier for the chain
                                        containing this turn.
            21-24    int    initSeqNum  Sequence number of initial residue in
                                        turn.
            25       string initICode   Insertion code of initial residue in
                                        turn.
            27-29    string endResName  Residue name of terminal residue of
                                        turn.
            31       string endChainId  Chain identifier for the chain
                                        containing this turn.
            32-35    int    endSeqNum   Sequence number of terminal residue of
                                        turn.
            36       string endICode    Insertion code of terminal residue of
                                        turn.
            41-70    string comment     Associated comment.
        """
        super(TURN, self).__init__(line)
        self.seq = int(string.strip(line[7:10]))
        self.turnId = string.strip(line[11:14])
        self.initResName = string.strip(line[15:18])
        self.initChainId = string.strip(line[19])
        self.initSeqNum = int(string.strip(line[20:24]))
        self.initICode = string.strip(line[24])
        self.endResName = string.strip(line[26:29])
        self.endChainId = string.strip(line[30])
        self.endSeqNum = int(string.strip(line[31:35]))
        self.endICode = string.strip(line[35])
        self.comment = string.strip(line[40:70])

@RegisterLineParser
class SHEET(BaseRecord):
    """ SHEET field

        SHEET records are used to identify the position of sheets in the
        molecule. Sheets are both named and numbered. The residues where the
        sheet begins and ends are noted.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD       DEFINITION
            -------------------------------------------------
             8 - 10  int    strand      Strand number which starts at 1 for
                                        each strand within a sheet and
                                        increases by one.
            12 - 14  string sheetID     Sheet identifier.
            15 - 16  int    numStrands  Number of strands in sheet.
            18 - 20  string initResName Residue name of initial residue.
            22       string initChainID Chain identifier of initial residue in
                                        strand.
            23 - 26  int    initSeqNum  Sequence number of initial residue in
                                        strand.
            27       string initICode   Insertion code of initial residue in
                                        strand.
            29 - 31  string endResName  Residue name of terminal residue.
            33       string endChainID  Chain identifier of terminal residue.
            34 - 37  int    endSeqNum   Sequence number of terminal residue.
            38       string endICode    Insertion code of terminal residue.
            39 - 40  int    sense       Sense of strand with respect to
                                        previous strand in the sheet. 0 if
                                        first strand, 1 if parallel, -1 if
                                        anti-parallel.
            42 - 45  string curAtom     Registration. Atom name in current
                                        strand.
            46 - 48  string curResName  Registration. Residue name in current
                                        strand.
            50       string curChainId  Registration. Chain identifier in
                                        current strand.
            51 - 54  int    curResSeq   Registration. Residue sequence number
                                        in current strand.
            55       string curICode    Registration. Insertion code in current
                                        strand.
            57 - 60  string prevAtom    Registration. Atom name in previous
                                        strand.
            61 - 63  string prevResName Registration. Residue name in previous
                                        strand.
            65       string prevChainId Registration. Chain identifier in
                                        previous strand.
            66 - 69  int    prevResSeq  Registration. Residue sequence number
                                        in previous strand.
            70       string prevICode   Registration. Insertion code in
                                        previous strand.
        """
        super(SHEET, self).__init__(line)
        self.strand = int(string.strip(line[7:10]))
        self.sheetID = string.strip(line[11:14])
        self.numStrands = int(string.strip(line[14:16]))
        self.initResName = string.strip(line[17:20])
        self.initChainID = string.strip(line[21])
        self.initSeqNum = int(string.strip(line[22:26]))
        self.initICode = string.strip(line[26])
        self.endResName = string.strip(line[28:31])
        self.endChainID = string.strip(line[32])
        self.endSeqNum = int(string.strip(line[33:37]))
        self.endICode = string.strip(line[37])
        self.sense = int(string.strip(line[38:40]))
        try:
            self.curAtom = string.strip(line[41:45])
            self.curResName = string.strip(line[45:48])
            self.curChainID = string.strip(line[49])
            try:  self.curResSeq = int(string.strip(line[50:54]))
            except ValueError:  self.curResSeq = None
            self.curICode = string.strip(line[54])
            self.prevAtom = string.strip(line[56:60])
            self.prevResName = string.strip(line[60:63])
            self.prevChainID = string.strip(line[64])
            try:  self.prevResSeq = int(string.strip(line[65:69]))
            except ValueError:  self.prevResSeq = None
            self.prevICode = string.strip(line[69])
        except IndexError:
            self.curAtom = None
            self.curResName = None
            self.curChainID = None
            self.curResSeq = None
            self.curICode = None
            self.prevAtom = None
            self.prevResName = None
            self.prevChainID = None
            self.prevResSeq = None
            self.prevICode = None

@RegisterLineParser
class HELIX(BaseRecord):
    """ HELIX field

        HELIX records are used to identify the position of helices in the
        molecule. Helices are both named and numbered. The residues where the
        helix begins and ends are noted, as well as the total length.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            ------------------------------------------------------
            8-10     int    serNum      Serial number of the helix.  This
                                        starts at 1 and increases
                                        incrementally.
            12-14    string helixID     Helix identifier.  In addition to a
                                        serial number, each helix is given an
                                        alphanumeric character helix identifier.
            16-18    string initResName Name of the initial residue.
            20       string initChainID Chain identifier for the chain
                                        containing this helix.
            22-25    int    initSeqNum  Sequence number of the initial residue.
            26       string initICode   Insertion code of the initial residue.
            28-30    string endResName  Name of the terminal residue of
                                        the helix.
            32       string endChainID  Chain identifier for the chain
                                        containing this helix.
            34-37    int    endSeqNum   Sequence number of the terminal residue.
            38       string endICode    Insertion code of the terminal residue.
            39-40    int    helixClass  Helix class (see below).
            41-70    string comment     Comment about this helix.
            72-76    int    length      Length of this helix.
        """
        super(HELIX, self).__init__(line)
        self.serNum = int(string.strip(line[7:10]))
        self.helixID = string.strip(line[11:14])
        self.initResName = string.strip(line[15:18])
        self.initChainID = string.strip(line[19])
        self.initSeqNum = int(string.strip(line[21:25]))
        self.initICode = string.strip(line[25])
        self.endResName = string.strip(line[27:30])
        self.endChainID = string.strip(line[31])
        self.endSeqNum = int(string.strip(line[33:37]))
        self.endICode = string.strip(line[37])
        try:  self.helixClass = int(string.strip(line[38:40]))
        except ValueError:  self.helixClass = None
        self.comment = string.strip(line[40:70])
        try:  self.length = int(string.strip(line[71:76]))
        except ValueError:  self.length = None

@RegisterLineParser
class FORMUL(BaseRecord):
    """ FORMUL field

        The FORMUL record presents the chemical formula and charge of a
        non-standard group.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD    DEFINITION
            -----------------------------------------------------
            9-10     int    compNum  Component number
            13-15    string hetID    Het identifier
            19       string asterisk * for water
            20-70    string text     Chemical formula
        """
        super(FORMUL, self).__init__(line)
        self.compNum = int(string.strip(line[8:10]))
        self.hetID = string.strip(line[12:15])
        self.asterisk = string.strip(line[19])
        self.text = string.strip(line[19:70])

@RegisterLineParser
class HETSYN(BaseRecord):
    """ HETSYN field

        This record provides synonyms, if any, for the compound in the
        corresponding (i.e., same hetID) HETNAM record. This is to allow
        greater flexibility in searching for HET groups.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD         DEFINITION
            -----------------------------------------------------
            12-14    string hetID         Het identifier, right-justified.
            16-70    string hetSynonyms   List of synonyms
        """
        super(HETSYN, self).__init__(line)
        self.hetID = string.strip(line[11:14])
        self.hetSynonyms = string.strip(line[15:70])

@RegisterLineParser
class HETNAM(BaseRecord):
    """ HETNAM field

        This record gives the chemical name of the compound with the given
        hetID.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            -----------------------------------------------------
            12-14    string hetID  Het identifier, right-justified.
            16-70    string text   Chemical name.
        """
        super(HETNAM, self).__init__(line)
        self.hetID = string.strip(line[11:14])
        self.text = string.strip(line[15:70])

@RegisterLineParser
class HET(BaseRecord):
    """ HET field

        HET records are used to describe non-standard residues, such as
        prosthetic groups, inhibitors, solvent molecules, and ions for which
        coordinates are supplied. Groups are considered HET if they are:
         - not one of the standard amino acids, and
         - not one of the nucleic acids (C, G, A, T, U, and I), and
         - not one of the modified versions of nucleic acids (+C, +G, +A, +T,
           +U, and +I), and
         - not an unknown amino acid or nucleic acid where UNK is used to
           indicate the unknown residue name.
        Het records also describe heterogens for which the chemical identity is
        unknown, in which case the group is assigned the hetID UNK.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD       DEFINITION
            --------------------------------------------------------
            8-10     string hetID       Het identifier, right-justified.
            13       string ChainID     Chain identifier.
            14-17    int    seqNum      Sequence number.
            18       string iCode       Insertion code.
            21-25    int    numHetAtoms Number of HETATM records for the
            31-70    string text        Text describing Het group.
        """
        super(HET, self).__init__(line)
        self.hetID = string.strip(line[7:10])
        self.chainID = string.strip(line[12])
        try:
            self.seqNum = int(string.strip(line[13]))
        except ValueError:
            self.seqNum = None
        self.iCode = string.strip(line[17])
        self.numHetAtoms = int(string.strip(line[20:25]))
        self.text = string.strip(line[30:70])

@RegisterLineParser
class MODRES(BaseRecord):
    """ MODRES field

        The MODRES record provides descriptions of modifications (e.g.,
        chemical or post-translational) to protein and nucleic acid residues.
        Included are a mapping between residue names given in a PDB entry and
        standard residues.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD   DEFINITION
            ---------------------------------------
            8-11     string idCode  ID code of this entry.
            13-15    string resName Residue name used in this entry.
            17       string chainID Chain identifier.
            19-22    int    seqNum  Sequence number.
            23       string iCode   Insertion code.
            25-27    string stdRes  Standard residue name.
            30-70    string comment Description of the residue modification.
        """
        super(MODRES, self).__init__(line)
        string.idCode = string.strip(line[7:11])
        string.resName = string.strip(line[12:15])
        string.chainID = string.strip(line[16])
        string.seqNum = int(string.strip(line[18:22]))
        string.iCode = string.strip(line[22])
        string.stdRes = string.strip(line[24:27])
        string.comment = string.strip(line[29:70])

@RegisterLineParser
class SEQRES(BaseRecord):
    """ SEQRES field

        SEQRES records contain the amino acid or nucleic acid sequence of
        residues in each chain of the macromolecule that was studied.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD   DEFINITION
            -----------------------------------------------------
            9-10     int    serNum  Serial number of the SEQRES record for the
                                    current chain.  Starts at 1 and increments
                                    by one each line.  Reset to 1 for each
                                    chain.
            12       string chainID Chain identifier.  This may be any single
                                    legal character, including a blank which is
                                    used if there is only one chain.
            14-17    int    numRes  Number of residues in the chain.  This
                                    value is repeated on every record.
            20-22    string resName Residue name.
            24-26    string resName Residue name.
            28-30    string resName Residue name.
            32-34    string resName Residue name.
            36-38    string resName Residue name.
            40-42    string resName Residue name.
            44-46    string resName Residue name.
            48-50    string resName Residue name.
            52-54    string resName Residue name.
            56-58    string resName Residue name.
            60-62    string resName Residue name.
            64-66    string resName Residue name.
            68-70    string resName Residue name.
        """
        super(SEQRES, self).__init__(line)
        self.serNum = int(string.strip(line[8:10]))
        self.chainID = string.strip(line[11])
        self.numRes = int(string.strip(line[13:17]))
        self.resName = []
        self.resName.append(string.strip(line[19:22]))
        self.resName.append(string.strip(line[23:26]))
        self.resName.append(string.strip(line[27:30]))
        self.resName.append(string.strip(line[31:34]))
        self.resName.append(string.strip(line[35:38]))
        self.resName.append(string.strip(line[39:42]))
        self.resName.append(string.strip(line[43:46]))
        self.resName.append(string.strip(line[47:50]))
        self.resName.append(string.strip(line[51:54]))
        self.resName.append(string.strip(line[55:58]))
        self.resName.append(string.strip(line[59:62]))
        self.resName.append(string.strip(line[63:66]))
        self.resName.append(string.strip(line[67:70]))

@RegisterLineParser
class SEQADV(BaseRecord):
    """ SEQADV field

        The SEQADV record identifies conflicts between sequence information in
        the ATOM records of the PDB entry and the sequence database entry given
        on DBREF. Please note that these records were designed to identify
        differences and not errors. No assumption is made as to which database
        contains the correct data. PDB may include REMARK records in the entry
        that reflect the depositor's view of which database has the correct
        sequence.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD    DEFINITION
            -----------------------------------------------------
            8-11     string idCode   ID code of this entry.
            13-15    string resName  Name of the PDB residue in conflict.
            17       string chainID  PDB chain identifier.
            19-22    int    seqNum   PDB sequence number.
            23       string iCode    PDB insertion code.
            25-28    string database Sequence database name.
            30-38    string dbIdCode Sequence database accession
                                     number.
            40-42    string dbRes    Sequence database residue name.
            44-48    int    dbSeq    Sequence database sequence number.
            50-70    string conflict Conflict comment.
        """
        super(SEQADV, self).__init__(line)
        self.idCode = string.strip(line[7:11])
        self.resName = string.strip(line[12:15])
        self.chainID = string.strip(line[16])
        try:  self.seqNum = int(string.strip(line[19:22]))
        except ValueError:  self.seqNum = None
        self.iCode = string.strip(line[22])
        self.database = string.strip(line[24:28])
        self.dbIdCode = string.strip(line[29:38])
        self.dbRes = string.strip(line[39:42])
        self.dbSeq = int(string.strip(line[43:48]))
        self.conflict = string.strip(line[49:70])

@RegisterLineParser
class DBREF(BaseRecord):
    """ DBREF field

        The DBREF record provides cross-reference links between PDB sequences
        and the corresponding database entry or entries. A cross reference to
        the sequence database is mandatory for each peptide chain with a length
        greater than ten (10) residues. For nucleic acid entries a DBREF record
        pointing to the Nucleic Acid Database (NDB) is mandatory when the
        corresponding entry exists in NDB.
    """

    def __init__(self, line):
        """
             Initialize by parsing a line.

             COLUMNS  TYPE   FIELD       DEFINITION
             ------------------------------------------------------
             8-11     string idCode      ID code of this entry.
             13       string chainID     Chain identifier.
             15-18    int    seqBegin    Initial sequence number of the PDB
                                         sequence segment.
             19       string insertBegin Initial insertion code of the PDB
                                         sequence segment.
             21-24    int    seqEnd      Ending sequence number of the PDB
                                         sequence segment.
             25       string insertEnd   Ending insertion code of the PDB
                                         sequence segment.
             27-32    string database    Sequence database name.  "PDB" when
                                         a corresponding sequence database
                                         entry has not been identified.
             34-41    string dbAccession Sequence database accession code.
                                         For GenBank entries, this is the
                                         NCBI gi number.
             43-54    string dbIdCode    Sequence database identification
                                         code.  For GenBank entries, this is
                                         the accession code.
             56-60    int    dbseqBegin  Initial sequence number of the
                                         database seqment.
             61       string dbinsBeg    Insertion code of initial residue
                                         of the segment, if PDB is the
                                         reference.
             63-67    int    dbseqEnd    Ending sequence number of the
                                         database segment.
             68       string dbinsEnd    Insertion code of the ending
                                         residue of the segment, if PDB is
                                         the reference.
        """
        super(DBREF, self).__init__(line)
        self.idCode = string.strip(line[7:11])
        self.chainID = string.strip(line[12])
        self.seqBegin = int(string.strip(line[14:18]))
        self.insertBegin = string.strip(line[18])
        self.seqEnd = int(string.strip(line[20:24]))
        self.insertEnd = string.strip(line[24])
        self.database = string.strip(line[26:32])
        self.dbAccession = string.strip(line[33:41])
        self.dbIdCode = string.strip(line[42:54])
        self.dbseqBegin = int(string.strip(line[55:60]))
        self.dbinsBeg = string.strip(line[60])
        self.dbseqEnd = int(string.strip(line[62:67]))
        try:  self.dbinsEnd = string.strip(line[67])
        except IndexError:  self.dbinsEnd = None

@RegisterLineParser
class REMARK(BaseRecord):
    """ REMARK field

        REMARK records present experimental details, annotations, comments, and
        information not included in other records. In a number of cases,
        REMARKs are used to expand the contents of other record types. A new
        level of structure is being used for some REMARK records. This is
        expected to facilitate searching and will assist in the conversion to a
        relational database.
    """

    def __init__(self, line):
        """
            Initialize by parsing line
        """
        super(REMARK, self).__init__(line)
        self.remarkNum = int(string.strip(line[7:10]))
        self.remarkDict = {}

        if self.remarkNum == 1:
            subfield = string.strip(line[11:20])
            if subfield == "REFERENCE":
                self.remarkDict["refNum"] = int(string.strip(line[21:70]))
            elif subfield == "AUTH":
                self.remarkDict["authorList"] = string.strip(line[19:70])
            elif subfield == "TITL":
                self.remarkDict["title"] = string.strip(line[19:70])
            elif subfield == "EDIT":
                self.remarkDict["editorList"] = string.strip(line[19:70])
            elif subfield == "REF":
                self.remarkDict["ref"] = string.strip(line[19:66])
            elif subfield == "PUBL":
                self.remarkDict["pub"] = string.strip(line[19:70])
            elif subfield == "REFN":
                self.remarkDict["refn"] = string.strip(line[19:70])
        elif self.remarkNum == 2:
            restr = string.strip(line[22:27])
            try:  self.remarkDict["resolution"] = float(restr)
            except ValueError:
                self.remarkDict["comment"] = string.strip(line[11:70])
        else:
            self.remarkDict["text"] = string.strip(line[11:70])



@RegisterLineParser
class JRNL(BaseRecord):
    """ JRNL field

        The JRNL record contains the primary literature citation that describes
        the experiment which resulted in the deposited coordinate set. There is
        at most one JRNL reference per entry. If there is no primary reference,
        then there is no JRNL reference. Other references are given in REMARK
        1.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD  DEFINITION
            -----------------------------------------------
            13-70    string text   See Details on web.
        """
        super(JRNL, self).__init__(line)
        #TODO: What is this mess?
        self.text = string.strip(line[12:70])

@RegisterLineParser
class SPRSDE(BaseRecord):
    """ SPRSDE field

        The SPRSDE records contain a list of the ID codes of entries that were
        made obsolete by the given coordinate entry and withdrawn from the PDB
        release set. One entry may replace many. It is PDB policy that only the
        principal investigator of a structure has the authority to withdraw it.
    """

    def __init__(self, line):
        """
            Initialize by parsing line

            COLUMNS  TYPE   FIELD      DEFINITION
            -----------------------------------------------
            12-20    string sprsdeDate Date this entry superseded the
                                       listed entries.
            22-25    string idCode     ID code of this entry.
            32-35    string sIdCode    ID code of a superseded entry.
            37-40    string sIdCode    ID code of a superseded entry.
            42-45    string sIdCode    ID code of a superseded entry.
            47-50    string sIdCode    ID code of a superseded entry.
            52-55    string sIdCode    ID code of a superseded entry.
            57-60    string sIdCode    ID code of a superseded entry.
            62-65    string sIdCode    ID code of a superseded entry.
            67-70    string sIdCode    ID code of a superseded entry.
        """
        super(SPRSDE, self).__init__(line)
        self.sprsdeDate = string.strip(line[11:20])
        self.idCode = string.strip(line[21:25])
        self.sIdCodes = []
        self.sIdCodes.append(string.strip(line[31:35]))
        self.sIdCodes.append(string.strip(line[36:40]))
        self.sIdCodes.append(string.strip(line[41:45]))
        self.sIdCodes.append(string.strip(line[46:50]))
        self.sIdCodes.append(string.strip(line[51:55]))
        self.sIdCodes.append(string.strip(line[56:60]))
        self.sIdCodes.append(string.strip(line[61:65]))
        self.sIdCodes.append(string.strip(line[66:70]))

@RegisterLineParser
class REVDAT(BaseRecord):
    """ REVDAT field

        REVDAT records contain a history of the modifications made to an entry
        since its release.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line.

            COLUMNS  TYPE   FIELD  DEFINITION
            -------------------------------------------------------
            8-10     int    modNum  Modification number.
            14-22    string modDate Date of modification (or release for
                                    new entries).
            24-28    string modId   Identifies this particular modification.
                                    It links to the archive used internally by
                                    PDB.
            32       int    modType An integer identifying the type of
                                    modification.  In case of revisions
                                    with more than one possible modType,
                                    the highest value applicable will be
                                    assigned.
            40-45    string record  Name of the modified record.
            47-52    string record  Name of the modified record.
            54-59    string record  Name of the modified record.
            61-66    string record  Name of the modified record.
        """
        super(REVDAT, self).__init__(line)
        self.modNum = int(string.strip(line[7:10]))
        self.modDate = string.strip(line[13:22])
        self.modId = string.strip(line[23:28])
        self.modType = int(string.strip(line[31]))
        self.records = []
        self.records.append(string.strip(line[39:45]))
        self.records.append(string.strip(line[46:52]))
        self.records.append(string.strip(line[53:59]))
        self.records.append(string.strip(line[60:66]))

@RegisterLineParser
class AUTHOR(BaseRecord):
    """ AUTHOR field

        The AUTHOR record contains the names of the people responsible for the
        contents of the entry.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD      DEFINITION
            --------------------------------------------------
            11-70    string authorList List of the author names, separated by
                                       commas
        """
        super(AUTHOR, self).__init__(line)
        self.authorList = string.strip(line[10:70])

@RegisterLineParser
class EXPDTA(BaseRecord):
    """ EXPDTA field

        The EXPDTA record identifies the experimental technique used. This may
        refer to the type of radiation and sample, or include the spectroscopic
        or modeling technique. Permitted values include:

        ELECTRON DIFFRACTION
        FIBER DIFFRACTION
        FLUORESCENCE TRANSFER
        NEUTRON DIFFRACTION
        NMR
        THEORETICAL MODEL
        X-RAY DIFFRACTION
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD     DEFINITION
            --------------------------------------------------
            11-70    string technique The experimental technique(s) with
                                      optional comment describing the sample
                                      or experiment
        """
        super(EXPDTA, self).__init__(line)
        self.technique = string.strip(line[10:70])

@RegisterLineParser
class KEYWDS(BaseRecord):
    """ KEYWDS field

        The KEYWDS record contains a set of terms relevant to the entry. Terms
        in the KEYWDS record provide a simple means of categorizing entries and
        may be used to generate index files. This record addresses some of the
        limitations found in the classification field of the HEADER record. It
        provides the opportunity to add further annotation to the entry in a
        concise and computer-searchable fashion.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD   DEFINITION
            --------------------------------------------------
            11-70    string keywds  Comma-separated list of keywords relevant
                                    to the entry
        """
        super(KEYWDS, self).__init__(line)
        self.keywds = string.strip(line[10:70])

@RegisterLineParser
class SOURCE(BaseRecord):
    """ SOURCE field

        The SOURCE record specifies the biological and/or chemical source of
        each biological molecule in the entry. Sources are described by both
        the common name and the scientific name, e.g., genus and species.
        Strain and/or cell-line for immortalized cells are given when they help
        to uniquely identify the biological entity studied.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD   DEFINITION
            --------------------------------------------------
            11-70    string source  Identifies the source of the macromolecule
                                    in a token: value format
        """
        super(SOURCE, self).__init__(line)
        self.source = string.strip(line[10:70])

@RegisterLineParser
class COMPND(BaseRecord):
    """ COMPND field

        The COMPND record describes the macromolecular contents of an entry.
        Each macromolecule found in the entry is described by a set of token:
        value pairs, and is referred to as a COMPND record component. Since the
        concept of a molecule is difficult to specify exactly, PDB staff may
        exercise editorial judgment in consultation with depositors in
        assigning these names.

        For each macromolecular component, the molecule name, synonyms, number
        assigned by the Enzyme Commission (EC), and other relevant details are
        specified.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line

            COLUMNS  TYPE   FIELD    DEFINITION
            --------------------------------------------------
            11-70    string compound Description of the molecular list
                                     components.
        """
        super(COMPND, self).__init__(line)
        self.compound = string.strip(line[10:70])

@RegisterLineParser
class CAVEAT(BaseRecord):
    """ CAVEAT field

        CAVEAT warns of severe errors in an entry. Use caution when using an
        entry containing this record.
    """

    def __init__(self, line):
        """
            Initialize by parsing line.

            COLUMNS  TYPE   FIELD   DEFINITION
            ----------------------------------------------------
            12-15    string idCode  PDB ID code of this entry.
            20-70    string comment Free text giving the reason for the
                                    CAVEAT.
        """
        super(CAVEAT, self).__init__(line)
        self.idCode = string.strip(line[11:15])
        self.comment = string.strip(line[19:70])

@RegisterLineParser
class TITLE(BaseRecord):
    """ TITLE field

        The TITLE record contains a title for the experiment or analysis that
        is represented in the entry. It should identify an entry in the PDB in
        the same way that a title identifies a paper.
    """

    def __init__(self, line):
        """
            Initialize by parsing a line.

            COLUMNS  TYPE   FIELD  DEFINITION
            ---------------------------------------------
            11-70    string title  Title of the experiment
        """
        super(TITLE, self).__init__(line)
        self.title = string.strip(line[10:70])

@RegisterLineParser
class OBSLTE(BaseRecord):
    """ OBSLTE field

        This record acts as a flag in an entry which has been withdrawn from
        the PDB's full release. It indicates which, if any, new entries have
        replaced the withdrawn entry.

        The format allows for the case of multiple new entries replacing one
        existing entry.
    """

    def __init__(self, line):
        """
           Initialize by parsing a line.

           COLUMNS  TYPE   FIELD    DEFINITION
           -----------------------------------------------
           12-20    string repDate  Date that this entry was replaced.
           22-25    string idCode   ID code of this entry.
           32-35    string rIdCode  ID code of entry that replaced
                                    this one.
           37-40    string rIdCode  ID code of entry that replaced
                                    this one.
           42-45    string rIdCode  ID code of entry that replaced
                                    this one.
           47-50    string rIdCode  ID code of entry that replaced
                                    this one.
           52-55    string rIdCode  ID code of entry that replaced
                                    this one.
           57-60    string rIdCode  ID code of entry that replaced
                                    this one.
           62-65    string rIdCode  ID code of entry that replaced
                                    this one.
           67-70    string rIdCode  ID code of entry that replaced
                                    this one.
        """
        super(OBSLTE, self).__init__(line)
        self.repDate = string.strip(line[11:20])
        self.idCode = string.strip(line[21:25])
        self.rIdCodes = []
        self.rIdCodes.append(string.strip(line[31:35]))
        self.rIdCodes.append(string.strip(line[36:40]))
        self.rIdCodes.append(string.strip(line[41:45]))
        self.rIdCodes.append(string.strip(line[46:50]))
        self.rIdCodes.append(string.strip(line[51:55]))
        self.rIdCodes.append(string.strip(line[56:60]))
        self.rIdCodes.append(string.strip(line[61:65]))
        self.rIdCodes.append(string.strip(line[67:70]))

@RegisterLineParser
class HEADER(BaseRecord):
    """ HEADER field

        The HEADER record uniquely identifies a PDB entry through the idCode
        field. This record also provides a classification for the entry.
        Finally, it contains the date the coordinates were deposited at the
        PDB. """

    def __init__(self, line):
        """
           Initialize by parsing a line.

           COLUMNS  TYPE   FIELD          DEFINITION
           ---------------------------------------------------------
           11-50    string classification Classifies the molecule(s)
           51-59    string depDate        Deposition date.  This is the date
                                          the coordinates were received by
                                          the PDB
           63-66    string idCode         This identifier is unique within PDB
        """
        super(HEADER, self).__init__(line)
        self.classification = string.strip(line[10:50])
        self.depDate = string.strip(line[50:59])
        self.IDcode = string.strip(line[62:66])

def readAtom(line):
    """
        If the ATOM/HETATM is not column-formatted, try to get some
        information by parsing whitespace from the right.  Look for
        five floating point numbers followed by the residue number.

        Parameters
            line:  The line to parse(string)
       if record == ATOM:
            self.serial = int(string.strip(line[6:11]))
            self.name = string.strip(line[12:16])
            self.altLoc = string.strip(line[16])
            self.resName = string.strip(line[17:20])
            self.chainID = string.strip(line[21])
            self.resSeq = int(string.strip(line[22:26]))
            self.iCode = string.strip(line[26])
            self.x = float(string.strip(line[30:38]))
            self.y = float(string.strip(line[38:46]))
            self.z = float(string.strip(line[46:54]))
            try:
                self.occupancy = float(string.strip(line[54:60]))
                self.tempFactor = float(string.strip(line[60:66]))
                self.segID = string.strip(line[72:76])
                self.element = string.strip(line[76:78])
                self.charge = string.strip(line[78:80])
            except ValueError, IndexError:
                self.occupancy = 0.00
                self.tempFactor = 0.00
                self.segID = 0
                self.element = 0
                self.charge = 0
        else: raise ValueError, record
    """

    # Try to find 5 consecutive floats
    words = string.split(line)
    size = len(words) - 1
    consec = 0
    for i in range(size):
        entry = words[size - i]
        try:
            val = float(entry)
            consec = consec + 1
            if consec == 5:
                break
        except ValueError:
            consec = 0

    record = string.strip(line[0:6])
    newline = line[0:22]
    newline = newline + string.rjust(words[size-i-1],4)
    newline = newline + string.rjust("",3)
    newline = newline + string.rjust(words[size-i],8)
    newline = newline + string.rjust(words[size-i+1],8)
    newline = newline + string.rjust(words[size-i+2],8)
    newline = newline + string.rjust(words[size-i+3],6)
    newline = newline + string.rjust(words[size-i+4],6)
    klass = lineParsers[record]
    obj = klass(newline)
    return obj

def readPDB(file):
    """ Parse PDB-format data into array of Atom objects.
        Parameters
          file:  open file object
        Returns (dict, errlist)
          dict:  a dictionary indexed by PDB record names
          errlist:  a list of record names that couldn't be parsed
    """

    pdblist = []  # Array of parsed lines (as objects)
    errlist = []  # List of records we can't parse

    #We can come up with nothing if can't get our file off the web.
    if file is None:
        return pdblist, errlist

    while 1:
        line = string.strip(file.readline())
        if line == '':
            break

        # We assume we have a method for each PDB record and can therefore
        # parse them automatically
        try:
            record = string.strip(line[0:6])
            if record not in errlist:
                klass = lineParsers[record]
                obj = klass(line)
                pdblist.append(obj)
        except KeyError as details:
            errlist.append(record)
            sys.stderr.write("Error parsing line: %s\n" % details)
            sys.stderr.write("<%s>\n" % string.strip(line))
            sys.stderr.write("Truncating remaining errors for record type:%s\n" % record)
        except StandardError as details:
            if record == "ATOM" or record == "HETATM":
                try:
                    obj = readAtom(line)
                    pdblist.append(obj)
                except StandardError, details:
                    sys.stderr.write("Error parsing line: %s\n" % details)
                    sys.stderr.write("<%s>\n" % string.strip(line))
            elif record == "SITE" or record == "TURN":
                pass
            elif record == "SSBOND" or record == "LINK":
                sys.stderr.write("Warning -- ignoring record: \n")
                sys.stderr.write("<%s>\n" % string.strip(line))
            else:
                sys.stderr.write("Error parsing line: %s\n" % details)
                sys.stderr.write("<%s>\n" % string.strip(line))

    return pdblist, errlist

def getRandom():
    """ Download a random PDB and return the path name.
        Returns
          path name of downloaded file
    """

    import os, random

    URL = "ftp://ftp.rcsb.org/pub/pdb/data/structures/all/pdb/"
    pdblines = os.popen("ncftpls %s" % URL).readlines()
    pdbline = string.join(pdblines)
    pdbline = string.replace(pdbline, "\n", "")
    pdbline = string.replace(pdbline, "@", "")
    pdbline = string.strip(pdbline)
    pdblist = string.split(pdbline)
    pdbZ = random.choice(pdblist)
    os.popen("ncftpget %s/%s" % (URL, pdbZ))
    os.popen("uncompress %s" % pdbZ)
    return pdbZ[:-2]

def main():
    """ Main driver for testing.  Parses set number of random PDBs """

    npdb = 1
    sys.stdout.write("Testing %d PDBs...\n" % npdb)
    for i in range(0, npdb):
        sys.stdout.write("Getting random PDB...\n")
        path = getRandom()
        sys.stdout.write("Parsing %s...\n" % path)
        pdbdict, errlist = readPDB(open(path, "rU"))
        if len(errlist) > 0:  sys.stdout.write("\tSkipped records: %s\n" % errlist)
        sys.stdout.write("\tNo skipped records.\n")

if __name__ == "__main__":  main()



"""
    Definitions for PDB2PQR

    This file contains classes associated with Amino Acid and Rotamer
    definitions as used by PDB2PQR.

    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

"""

__date__ = "15 May 2008"
__author__ = "Jens Erik Nielsen, Todd Dolinsky, Yong Huang"

AAPATH = "dat/AA.xml"
NAPATH = "dat/NA.xml"
PATCHPATH = "dat/PATCHES.xml"

import os
import copy
import re
from xml import sax
from pdb import *
from utilities import *
from structures import *
from routines import *

from errors import PDBInternalError


class DefinitionHandler(sax.ContentHandler):
    def __init__(self):
        self.curelement = ""
        self.curatom = None
        self.curholder = None
        self.curobj = None
        self.map = {}
        self.patches = []
        return

    def startElement(self, name, attributes):
        if name == "residue":
            obj = DefinitionResidue()
            self.curholder = obj
            self.curobj = obj
        elif name == "patch":
            obj = Patch()
            self.curholder = obj
            self.curobj = obj
        elif name == "atom":
            obj = DefinitionAtom()
            self.curatom = obj
            self.curobj = obj
        else:
            self.curelement = name
        return

    def endElement(self, name):
        if name == "residue":  # Complete Residue object
            residue = self.curholder
            if not isinstance(residue, DefinitionResidue):
                raise PDBInternalError("Internal error parsing XML!")
            resname = residue.name
            if resname == "":
                raise PDBInternalError("Residue name not set in XML!")
            else:
                self.map[resname] = residue
                self.curholder = None
                self.curobj = None

        elif name == "patch":  # Complete patch object
            patch = self.curholder
            if not isinstance(patch, Patch):
                raise PDBInternalError("Internal error parsing XML!")
            patchname = patch.name
            if patchname == "":
                raise PDBInternalError("Residue name not set in XML!")
            else:
                self.patches.append(patch)
                self.curholder = None
                self.curobj = None

        elif name == "atom":  # Complete atom object
            atom = self.curatom
            if not isinstance(atom, DefinitionAtom):
                raise PDBInternalError("Internal error parsing XML!")
            atomname = atom.name
            if atomname == "":
                raise PDBInternalError("Atom name not set in XML!")
            else:
                self.curholder.map[atomname] = atom
                self.curatom = None
                self.curobj = self.curholder

        else:  # Just free the current element namespace
            self.curelement = ""

        return self.map

    def characters(self, text):
        if text.isspace():
            return

        # If this is a float, make it so
        try:
            value = float(str(text))
        except ValueError:
            value = str(text)

        # Special cases - lists and dictionaries
        if self.curelement == "bond":
            self.curobj.bonds.append(value)
        elif self.curelement == "dihedral":
            self.curobj.dihedrals.append(value)
        elif self.curelement == "altname":
            self.curholder.altnames[value] = self.curatom.name
        elif self.curelement == "remove":
            self.curobj.remove.append(value)
        else:
            setattr(self.curobj, self.curelement, value)
        return


class Definition:
    """
    Definition class

    The Definition class contains the structured definitions found
    in the files and several mappings for easy access to the information.
    """

    def __init__(self):
        """
        Create a new Definition Object
        """
        self.map = {}
        self.patches = {}

        handler = DefinitionHandler()
        sax.make_parser()

        for path in [AAPATH, NAPATH]:
            defpath = getDatFile(path)
            if defpath == "":
                raise PDBInternalError("%s not found!" % path)

            acidFile = open(defpath)
            sax.parseString(acidFile.read(), handler)
            acidFile.close()

            self.map.update(handler.map)

        # Now handle patches

        defpath = getDatFile(PATCHPATH)
        if defpath == "":
            raise PDBInternalError("%s not found!" % PATCHPATH)

        handler.map = {}
        patchFile = open(defpath)
        sax.parseString(patchFile.read(), handler)
        patchFile.close()

        # Apply specific patches to the reference object, allowing users
        #  to specify protonation states in the PDB file

        for patch in handler.patches:
            if patch.newname != "":

                # Find all residues matching applyto

                resnames = self.map.keys()
                for name in resnames:
                    regexp = re.compile(patch.applyto).match(name)
                    if not regexp:
                        continue
                    newname = patch.newname.replace("*", name)
                    self.addPatch(patch, name, newname)

            # Either way, make sure the main patch name is available

            self.addPatch(patch, patch.applyto, patch.name)

    def addPatch(self, patch, refname, newname):
        """
        Add a patch to a definition residue.

        Parameters
            patch:  The patch object to add (Patch)
            refname:  The name of the object to add the patch to (string)
            newname:  The name of the new (patched) object (string)
        """
        try:
            aadef = self.map[refname]  # The reference
            patchResidue = copy.deepcopy(aadef)

            # Add atoms from patch

            for atomname in patch.map:
                patchResidue.map[atomname] = patch.map[atomname]
                for bond in patch.map[atomname].bonds:
                    if bond not in patchResidue.map:
                        continue
                    if atomname not in patchResidue.map[bond].bonds:
                        patchResidue.map[bond].bonds.append(atomname)

            # Rename atoms as directed

            for key in patch.altnames:
                patchResidue.altnames[key] = patch.altnames[key]

            # Remove atoms as directed

            for remove in patch.remove:
                if not patchResidue.hasAtom(remove):
                    continue
                removebonds = patchResidue.map[remove].bonds
                del patchResidue.map[remove]
                for bond in removebonds:
                    if remove in patchResidue.map[bond].bonds:
                        patchResidue.map[bond].bonds.remove(remove)

            # Add the new dihedrals

            for dihedral in patch.dihedrals:
                patchResidue.dihedrals.append(dihedral)

            # Point at the new reference

            self.map[newname] = patchResidue

            # Store the patch

            self.patches[newname] = patch

        except KeyError:  # Just store the patch
            self.patches[newname] = patch


class Patch:
    """
    Patch the definitionResidue class
    """

    def __init__(self):
        """
        Initialize the Patch object.
        """
        self.name = ""
        self.applyto = ""
        self.map = {}
        self.remove = []
        self.altnames = {}
        self.dihedrals = []
        self.newname = ""

    def __str__(self):
        """
        A basic string representation for debugging
        """
        text = "%s\n" % self.name
        text += "Apply to: %s\n" % self.applyto
        text += "Atoms to add: \n"
        for atom in self.map:
            text += "\t%s\n" % str(self.map[atom])
        text += "Atoms to remove: \n"
        for remove in self.remove:
            text += "\t%s\n" % remove
        text += "Alternate naming map: \n"
        text += "\t%s\n" % self.altnames
        return text


class DefinitionResidue(Residue):
    """
    DefinitionResidue class

    The DefinitionResidue class extends the Residue class to allow for a
    trimmed down initializing function.
    """

    def __init__(self):
        """
        Initialize the class using a few parameters

        Parameters:
            name: The abbreviated amino acid name of the DefinitionResidue
        """
        self.name = ""
        self.dihedrals = []
        self.map = {}
        self.altnames = {}

    def __str__(self):
        """
        A basic string representation for debugging
        """
        text = "%s\n" % self.name
        text += "Atoms: \n"
        for atom in self.map:
            text += "\t%s\n" % str(self.map[atom])
        text += "Dihedrals: \n"
        for dihedral in self.dihedrals:
            text += "\t%s\n" % dihedral
        text += "Alternate naming map: \n"
        text += "\t%s\n" % self.altnames
        return text

    def addDihedral(self, atom):
        """
        Add the atom to the list of dihedral bonds

        Parameters:
            atom: The atom to be added
        """
        self.dihedralatoms.append(atom)

    def getNearestBonds(self, atomname):
        """
        Parameters
            number:   The number of bonds to get
        Returns
            bonds:    A list of atomnames that are within three bonds of
                      the atom and present in residue (list)
        """
        bonds = []
        lev2bonds = []
        atom = self.map[atomname]

        # Get directly bonded (length = 1) atoms

        for bondedatom in atom.bonds:
            if bondedatom not in bonds:
                bonds.append(bondedatom)

        # Get bonded atoms 2 bond lengths away

        for bondedatom in atom.bonds:
            for bond2 in self.map[bondedatom].bonds:
                if bond2 not in bonds and bond2 != atomname:
                    bonds.append(bond2)
                    lev2bonds.append(bond2)

        # Get bonded atoms 3 bond lengths away

        for lev2atom in lev2bonds:
            for bond3 in self.map[lev2atom].bonds:
                if bond3 not in bonds:
                    bonds.append(bond3)

        return bonds


class DefinitionAtom(Atom):
    """
    A trimmed down version of the Atom class
    """

    def __init__(self, name=None, x=None, y=None, z=None):
        """
        Initialize the class
        """
        self.name = name
        self.x = x
        self.y = y
        self.z = z
        if name == None:
            self.name = ""
        if x == None:
            self.x = 0.0
        if y == None:
            self.y = 0.0
        if z == None:
            self.z = 0.0
        self.bonds = []

    def __str__(self):
        """
        A basic string representation for debugging
        """
        text = "%s: %.3f %.3f %.3f" % (self.name, self.x, self.y, self.z)
        for bond in self.bonds:
            text += " %s" % bond
        return text

    def isBackbone(self):
        """
        Return true if atom name is in backbone, otherwise false

        Returns
            state: 1 if true, 0 if false
        """
        if self.name in BACKBONE:
            return 1
        else:
            return 0



def new_pdb_line(aID, aname, resname, resnumb, x, y, z, chain=" "):
    pdb_format = "ATOM  {:5d} {:4s} {:4s}{:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}\n"
    return pdb_format.format(aID, aname, resname, chain, resnumb, x, y, z)


def new_pqr_line(aID, aname, resname, resnumb, x, y, z, charge, radius, chain=" "):
    pdb_format = (
        "ATOM  {:5d} {:4s} {:4s}{:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}{:8.3f}{:8.3f}\n"
    )
    return pdb_format.format(
        aID, aname, resname, chain, resnumb, x, y, z, charge, radius
    )


def new_gro_line(aID, aname, resname, resnumb, x, y, z):
    gro_format = "{:5}{:5}{:>5}{:5}{:8.3f}{:8.3f}{:8.3f}\n"
    return gro_format.format(resnumb, resname, aname, aID, x, y, z)


def correct_longHs(aname):
    if aname[1] == "H" and aname[0] in "12" and aname[3] in "12":
        return aname[1:3] + aname[3] + aname[0]
    return aname


def read_pqr_line(line):
    (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
    charge = float(line[54:62])
    radius = float(line[62:70])
    return (aname, anumb, resname, chain, resnumb, x, y, z, charge, radius)


def read_pdb_line(line):
    aname = line[12:16].strip()
    anumb = int(line[5:11].strip())
    resname = line[17:21].strip()
    chain = line[21]
    resnumb = int(line[22:26])
    x = float(line[30:38])
    y = float(line[38:46])
    z = float(line[46:54])
    if len(aname) == 4:
        aname = correct_longHs(aname)
    return (aname, anumb, resname, chain, resnumb, x, y, z)


def read_gro_line(line):
    resnumb = int(line[:5].strip())
    resname = line[5:10].strip()
    aname = line[10:15].strip()
    anumb = int(line[15:20].strip())
    x = float(line[20:28].strip())
    y = float(line[28:36].strip())
    z = float(line[36:44].strip())
    return (aname, anumb, resname, resnumb, x, y, z)


def gro2pdb(f_in, f_out):
    new_pdb_content = "REMARK    CONVERTED from {} by PypKa\n".format(f_in)
    with open(f_in) as f:
        line_counter = 0
        natoms_left = 0
        natoms = 0
        lines = f.read().splitlines()
        last_line = lines[-1]
        box = [float(i) * 10 for i in last_line.split()[:3]]
        new_pdb_content += (
            "CRYST1  {0:3.3f}  {1:3.3f}  {2:3.3f}  "
            "60.00  60.00  90.00 P 1           1\n".format(box[0], box[1], box[2])
        )
        for line in lines:
            line_counter += 1
            if natoms_left > 0:
                natoms_left -= 1
                natoms += 1
                (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line)

                x, y, z = x * 10, y * 10, z * 10

                new_pdb_content += new_pdb_line(
                    anumb, aname, resname, resnumb, x, y, z, chain="A"
                )
            elif line_counter == 2:
                natoms_left = int(line.strip())
    new_pdb_content += "TER\nENDMDL\n"
    with open(f_out, "w") as f_new:
        f_new.write(new_pdb_content)


def get_grobox_size(f_in):
    with open(f_in) as f:
        lines = f.read().splitlines()
        last_line = lines[-1]
        box = [float(i) * 10 for i in last_line.split()[:3]]
    return box


def get_chains_from_file(f_in):
    chain_list = []
    with open(f_in) as f:
        for line in f:
            if "ATOM " == line[0:5]:
                pdb_cols = read_pdb_line(line)
                chain = pdb_cols[3]
                if chain not in chain_list:
                    chain_list.append(chain)
    return chain_list


#!/usr/bin/python

""" psize class

    Get dimensions and other information from a PQR file.

    Originally written by Dave Sept
    Additional APBS-specific features added by Nathan Baker
    Ported to Python/Psize class by Todd Dolinsky and subsequently
    hacked by Nathan Baker

        ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

__date__ = "4 June 2008"
__author__ = "Dave Sept, Nathan Baker, Todd Dolinsky, Yong Huang"

import string, sys, getopt
from sys import stdout, stderr
from math import log

class Psize:
    """Master class for parsing input files and suggesting settings"""
    def __init__(self):
        self.constants = {"cfac": 1.7, "fadd":20, "space": 0.50, "gmemfac": 200, "gmemceil": 400, "ofrac":0.1, "redfac": 0.25 }
        self.minlen = [None, None, None]
        self.maxlen = [None, None, None]
        self.q = 0.0
        self.gotatom = 0
        self.gothet = 0
        self.olen = [0.0, 0.0, 0.0]
        self.cen = [0.0, 0.0, 0.0]
        self.clen = [0.0, 0.0, 0.0]
        self.flen = [0.0, 0.0, 0.0]
        self.n = [0, 0, 0]
        self.np = [0.0, 0.0, 0.0]
        self.nsmall = [0,0,0]
        self.nfocus = 0

    def parseString(self, structure):
        """ Parse the input structure as a string in PDB or PQR format """
        lines = string.split(structure, "\n")
        self.parseLines(lines)

    def parseInput(self, filename):
        """ Parse input structure file in PDB or PQR format """
        with open(filename, 'rU') as f:
            self.parseLines(f.readlines())

    def parseLines(self, lines):
        """ Parse the lines """
        for line in lines:
            if string.find(line,"ATOM") == 0:
                subline = string.replace(line[30:], "-", " -")
                words = string.split(subline)
                if len(words) < 5:    
                    continue
                self.gotatom += 1
                self.q = self.q + float(words[3])
                rad = float(words[4])
                center = []
                for word in words[0:3]:
                    center.append(float(word))
                for i in range(3):
                    if self.minlen[i] == None or center[i]-rad < self.minlen[i]:
                        self.minlen[i] = center[i]-rad
                    if self.maxlen[i] == None or center[i]+rad > self.maxlen[i]:
                        self.maxlen[i] = center[i]+rad
            elif string.find(line, "HETATM") == 0:
                self.gothet = self.gothet + 1
                # Special handling for no ATOM entries in the pqr file, only HETATM entries
                if self.gotatom == 0:
                    subline = string.replace(line[30:], "-", " -")
                    words = string.split(subline)
                    if len(words) < 5:    
                        continue
                    self.q = self.q + float(words[3])
                    rad = float(words[4])
                    center = []
                    for word in words[0:3]:
                        center.append(float(word))
                    for i in range(3):
                        if self.minlen[i] == None or center[i]-rad < self.minlen[i]:
                            self.minlen[i] = center[i]-rad
                        if self.maxlen[i] == None or center[i]+rad > self.maxlen[i]:
                            self.maxlen[i] = center[i]+rad
    
    def setConstant(self, name, value):
        """ Set a constant to a value; returns 0 if constant not found """
        try:
            self.constants[name] = value
            return 1
        except KeyError:
            return 0

    def getConstant(self, name):
        """ Get a constant value; raises KeyError if constant not found """
        return self.constants[name]

    def setLength(self, maxlen, minlen):
        """ Compute molecule dimensions """
        for i in range(3):
            self.olen[i] = maxlen[i] - minlen[i]
            if self.olen[i] < 0.1:
                self.olen[i] = 0.1
        return self.olen


    def setCoarseGridDims(self, olen):
        """ Compute coarse mesh dimensions """
        for i in range(3):
            self.clen[i] = self.constants["cfac"] * olen[i]
        return self.clen

    def setFineGridDims(self, olen, clen):
        """ Compute fine mesh dimensions """
        for i in range(3):
            self.flen[i] = olen[i] + self.constants["fadd"]
            if self.flen[i] > clen[i]:
                #str = "WARNING: Fine length (%.2f) cannot be larger than coarse length (%.2f)\n" % (self.flen[i], clen[i])
                #str = str + "         Setting fine grid length equal to coarse grid length\n"
                #stdout.write(str)
                self.flen[i] = clen[i]
        return self.flen

    def setCenter(self, maxlen, minlen):
        """ Compute molecule center """
        for i in range(3):
            self.cen[i] = (maxlen[i] + minlen[i]) / 2
        return self.cen


    def setFineGridPoints(self, flen):
        """ Compute mesh grid points, assuming 4 levels in MG hierarchy """
        tn = [0,0,0]
        for i in range(3):
            tn[i] = int(flen[i]/self.constants["space"] + 0.5)
            self.n[i] = 32*(int((tn[i] - 1) / 32.0 + 0.5)) + 1
            if self.n[i] < 33:
                self.n[i] = 33
        return self.n

    def setSmallest(self, n):
        """ Compute parallel division in case memory requirement above ceiling
        Find the smallest dimension and see if the number of grid points in
        that dimension will fit below the memory ceiling
        Reduce nsmall until an nsmall^3 domain will fit into memory """
        nsmall = []
        for i in range(3):
            nsmall.append(n[i])
        while 1:
            nsmem = 200.0 * nsmall[0] * nsmall[1] * nsmall[2] / 1024 / 1024
            if nsmem < self.constants["gmemceil"]: break
            else:
                i = nsmall.index(max(nsmall))
                nsmall[i] = 32 * ((nsmall[i] - 1)/32 - 1) + 1
                if nsmall <= 0:
                    stdout.write("You picked a memory ceiling that is too small\n")
                    sys.exit(0)        

        self.nsmall = nsmall
        return nsmall

    def setProcGrid(self, n, nsmall):
        """ Calculate the number of processors required to span each 
        dimension """

        zofac = 1 + 2 * self.constants["ofrac"]
        for i in range(3):
            self.np[i] = n[i]/float(nsmall[i])
            if self.np[i] > 1: self.np[i] = int(zofac*n[1]/nsmall[i] + 1.0)
        return self.np
                                                
    def setFocus(self, flen, np, clen):
        """ Calculate the number of levels of focusing required for each
        processor subdomain """

        nfoc = [0,0,0]
        for i in range(3):
            nfoc[i] = int(log((flen[i]/np[i])/clen[i])/log(self.constants["redfac"]) + 1.0)
        nfocus = nfoc[0]
        if nfoc[1] > nfocus: nfocus = nfoc[1]
        if nfoc[2] > nfocus: nfocus = nfoc[2]
        if nfocus > 0: nfocus = nfocus + 1
        self.nfocus = nfocus

    def setAll(self):
        """ Set up all of the things calculated individually above """
        maxlen = self.getMax()
        minlen = self.getMin()
        self.setLength(maxlen, minlen)
        olen = self.getLength()
        
        self.setCoarseGridDims(olen)
        clen = self.getCoarseGridDims()        
        
        self.setFineGridDims(olen, clen)
        flen = self.getFineGridDims()
        
        self.setCenter(maxlen, minlen)
        cen = self.getCenter()
        
        self.setFineGridPoints(flen)
        n = self.getFineGridPoints()
        
        self.setSmallest(n)
        nsmall = self.getSmallest()
        
        self.setProcGrid(n, nsmall)
        np = self.getProcGrid()
        
        self.setFocus(flen, np, clen)
        nfocus = self.getFocus()
        
    def getMax(self): return self.maxlen
    def getMin(self): return self.minlen
    def getCharge(self): return self.q
    def getLength(self): return self.olen
    def getCoarseGridDims(self): return self.clen
    def getFineGridDims(self): return self.flen
    def getCenter(self): return self.cen
    def getFineGridPoints(self): return self.n
    def getSmallest(self): return self.nsmall
    def getProcGrid(self): return self.np
    def getFocus(self): return self.nfocus

    def runPsize(self, filename):
        """ Parse input PQR file and set parameters """
        self.parseInput(filename)
        self.setAll()
    
    def printResults(self):
        """ Return a string with the formatted results """

        str = "\n"
        
        if self.gotatom > 0:

            maxlen = self.getMax()
            minlen = self.getMin()
            q = self.getCharge()
            olen = self.getLength()
            clen = self.getCoarseGridDims()        
            flen = self.getFineGridDims()
            cen = self.getCenter()
            n = self.getFineGridPoints()
            nsmall = self.getSmallest()
            np = self.getProcGrid()
            nfocus = self.getFocus()

            # Compute memory requirements

            nsmem = 200.0 * nsmall[0] * nsmall[1] * nsmall[2] / 1024 / 1024
            gmem = 200.0 * n[0] * n[1] * n[2] / 1024 / 1024
            
            # Print the calculated entries
            str = str + "################# MOLECULE INFO ####################\n"
            str = str + "Number of ATOM entries = %i\n" % self.gotatom
            str = str + "Number of HETATM entries (ignored) = %i\n" % self.gothet
            str = str + "Total charge = %.3f e\n" % q
            str = str + "Dimensions = %.3f x %.3f x %.3f A\n" % (olen[0], olen[1], olen[2])
            str = str + "Center = %.3f x %.3f x %.3f A\n" % (cen[0], cen[1], cen[2])
            str = str + "Lower corner = %.3f x %.3f x %.3f A\n" % (minlen[0], minlen[1], minlen[2])
            str = str + "Upper corner = %.3f x %.3f x %.3f A\n" % (maxlen[0], maxlen[1], maxlen[2])

            str = str + "\n"
            str = str + "############## GENERAL CALCULATION INFO #############\n"
            str = str + "Coarse grid dims = %.3f x %.3f x %.3f A\n" % (clen[0],
clen[1], clen[2])
            str = str + "Fine grid dims = %.3f x %.3f x %.3f A\n" % (flen[0], flen[1], flen[2])
            str = str + "Num. fine grid pts. = %i x %i x %i\n" % (n[0], n[1], n[2])

            if gmem > self.constants["gmemceil"]:
                str = str + "Parallel solve required (%.3f MB > %.3f MB)\n" % (gmem, self.constants["gmemceil"])
                str = str + "Total processors required = %i\n" % (np[0]*np[1]*np[2])
                str = str + "Proc. grid = %i x %i x %i\n" % (np[0], np[1], np[2])
                str = str + "Grid pts. on each proc. = %i x %i x %i\n" % (nsmall[0], nsmall[1], nsmall[2])
                xglob = np[0]*round(nsmall[0]/(1 + 2*self.constants["ofrac"]) - .001)
                yglob = np[1]*round(nsmall[1]/(1 + 2*self.constants["ofrac"]) - .001)
                zglob = np[2]*round(nsmall[2]/(1 + 2*self.constants["ofrac"]) - .001)
                if np[0] == 1: xglob = nsmall[0]
                if np[1] == 1: yglob = nsmall[1]
                if np[2] == 1: zglob = nsmall[2]
                str = str + "Fine mesh spacing = %g x %g x %g A\n" % (flen[0]/(xglob-1), flen[1]/(yglob-1), flen[2]/(zglob-1))
                str = str + "Estimated mem. required for parallel solve = %.3f MB/proc.\n" % nsmem
                ntot = nsmall[0]*nsmall[1]*nsmall[2]

            else:
                str = str + "Fine mesh spacing = %g x %g x %g A\n" % (flen[0]/(n[0]-1), flen[1]/(n[1]-1), flen[2]/(n[2]-1))
                str = str + "Estimated mem. required for sequential solve = %.3f MB\n" % gmem
                ntot = n[0]*n[1]*n[2]

            str = str + "Number of focusing operations = %i\n" % nfocus

            str = str + "\n"
            str = str + "################# ESTIMATED REQUIREMENTS ####################\n"
            str = str + "Memory per processor                   = %.3f MB\n" % (200.0*ntot/1024/1024)
            str = str + "Grid storage requirements (ASCII)      = %.3f MB\n" % (8.0*12*np[0]*np[1]*np[2]*ntot/1024/1024)
            str = str + "\n"

        else:
            str = str + "No ATOM entires in file!\n\n"

        return str

def usage(rc):
    """ Print usage information and exit with error code rc """
    psize = Psize()
    usage = "\n"
    usage = usage + "Psize script\n"
    usage = usage + "Usage: psize.py [opts] <filename>\n"
    usage = usage + "Optional Arguments:\n"
    usage = usage + "  --help               : Display this text\n"
    usage = usage + "  --cfac=<value>       : Factor by which to expand mol dims to\n"
    usage = usage + "                         get coarse grid dims\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("cfac")
    usage = usage + "  --fadd=<value>       : Amount to add to mol dims to get fine\n"
    usage = usage + "                         grid dims\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("fadd")
    usage = usage + "  --space=<value>      : Desired fine mesh resolution\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("space")
    usage = usage + "  --gmemfac=<value>    : Number of bytes per grid point required\n"
    usage = usage + "                         for sequential MG calculation\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("gmemfac")
    usage = usage + "  --gmemceil=<value>   : Max MB allowed for sequential MG\n"
    usage = usage + "                         calculation.  Adjust this to force the\n"
    usage = usage + "                         script to perform faster calculations (which\n"
    usage = usage + "                         require more parallelism).\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("gmemceil")
    usage = usage + "  --ofrac=<value>       : Overlap factor between mesh partitions\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("ofrac")
    usage = usage + "  --redfac=<value>     : The maximum factor by which a domain\n"
    usage = usage + "                         dimension can be reduced during focusing\n"
    usage = usage + "                         [default = %g]\n" % psize.getConstant("redfac")

    
    stderr.write(usage)
    sys.exit(rc)

def main():
    filename = ""
    shortOptList = "h"
    longOptList = ["help", "cfac=", "fadd=", "space=", "gmemfac=", "gmemceil=", "ofrac=", "redfac="]
    try:
        opts, args = getopt.getopt(sys.argv[1:], shortOptList, longOptList)
    except getopt.GetoptError, details:
        stderr.write("Option error (%s)!\n" % details)
        usage(2)
    if len(args) != 1: 
        stderr.write("Invalid argument list!\n")
        usage(2)
    else:
        filename = args[0]

    psize = Psize()    

    for o, a in opts:
        if o.lower() == "--help" or o == "-h":
            usage(0)
        if o.lower() == "--cfac":
            psize.setConstant("cfac", float(a))
        if o.lower() == "--fadd":
            psize.setConstant("fadd", int(a))
        if o.lower() == "--space":
            psize.setConstant("space", float(a))
        if o.lower() == "--gmemfac":
            psize.setConstant("gmemfac", int(a))
        if o.lower() == "--gmemceil":
            psize.setConstant("gmemceil",  int(a))
        if o.lower() == "--ofrac":
            psize.setConstant("ofrac", float(a))
        if o.lower() == "--redfac":
            psize.setConstant("redfac", float(a))

    psize.runPsize(filename)
    
    stdout.write("# Constants used: \n");
    for key in psize.constants.keys():
        stdout.write("# \t%s: %s\n" % (key, psize.constants[key]))
    stdout.write("# Run:\n")
    stdout.write("#    `%s --help`\n" % sys.argv[0])
    stdout.write("# for more information on these default values\n" )
    stdout.write(psize.printResults())

    
if __name__ == "__main__": main()



"""
    Utilities for PDB2PQR Suite

    This module provides various utilities for the PDB2PQR suite to be
    imported into other Python scripts.
    
    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

__date__ = "6 November 2007"
__author__ = "Todd Dolinsky, Yong Huang"
SMALL = 1.0e-7
DIHEDRAL = 57.2958

import math
import os
from os.path import splitext
import sys
from aconf import INSTALLDIR, TMPDIR


def startLogFile(jobName, fileName, logInput):
    with open("%s%s%s/%s" % (INSTALLDIR, TMPDIR, jobName, fileName), "w") as f:
        f.write(logInput)


def appendToLogFile(jobName, fileName, logInput):
    with open("%s%s%s/%s" % (INSTALLDIR, TMPDIR, jobName, fileName), "a") as f:
        f.write(logInput)


def resetLogFile(jobName, fileName):
    """
    For clearing out old log files if needed.
    Used mainly for removing apbs_end_time if apbs is rerun.
    """
    filename = "%s%s%s/%s" % (INSTALLDIR, TMPDIR, jobName, fileName)
    try:
        os.remove(filename)
    except EnvironmentError:
        pass


def getTrackingScriptString(jobid=None):
    """
    For injecting tracking script into a web page.

    jobid -> current jobid. Adds "jobid" custom variable to events and page views on this page.
    """
    customVarString = ""

    if jobid is not None:
        customVarString = "_gaq.push(['_setCustomVar',1,'jobid','{jobid}',3]);".format(
            jobid=str(jobid)
        )

    # If you look closely you'll see escaped { and }.
    string = """<script type="text/javascript">

  var _gaq = _gaq || [];
  _gaq.push(['_setAccount', 'UA-11026338-3']);
  _gaq.push(['_setDomainName', 'none']);
  _gaq.push(['_setAllowLinker', true]);
  {customVar}
  _gaq.push(['_trackPageview']);

  (function() {{
    var ga = document.createElement('script'); ga.type = 'text/javascript'; ga.async = true;
    ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
    var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(ga, s);
  }})();

</script>""".format(
        customVar=customVarString
    )
    return string


def getEventTrackingString(category, action, label, value=None):
    valueString = ", {value}".format(value=value) if value is not None else ""
    eventString = (
        '_gaq.push(["_trackEvent", "{category}", "{action}", "{label}"{valuestr}]);\n'
    )
    return eventString.format(
        category=str(category),
        action=str(action),
        label=str(label),
        valuestr=valueString,
    )


class ExtraOptions(object):
    pass


def createPropkaOptions(pH, verbose=False, reference="neutral"):
    """
    Create a propka options object for running propka.
    """
    # build propka options
    propkaOpts = ExtraOptions()
    propkaOpts.pH = pH
    propkaOpts.reference = reference
    propkaOpts.chains = None
    propkaOpts.thermophiles = None
    propkaOpts.alignment = None
    propkaOpts.mutations = None
    propkaOpts.verbose = verbose
    propkaOpts.protonation = "old-school"
    propkaOpts.window = (0.0, 14.0, 1.0)
    propkaOpts.grid = (0.0, 14.0, 0.1)
    propkaOpts.mutator = None
    propkaOpts.mutator_options = None
    propkaOpts.display_coupled_residues = None
    propkaOpts.print_iterations = None
    propkaOpts.version_label = "Nov30"

    # These adds a few bits to propkaOpts
    from propka30.Source import lib

    lib.interpretMutator(propkaOpts)
    # With the current defaults used here this does not do anything.
    # However if we start adding the propka options we'll need to do this.
    lib.setDefaultAlignmentFiles(propkaOpts)

    return propkaOpts


# totalDelay = 0
#
# def getEventTrackingString(category, action, label, value=None):
#    global totalDelay
#    valueString = ', "{value}"'.format(value=value) if value is not None else ""
#    eventString = 'setTimeout(\'_gaq.push(["_trackEvent", "{category}", "{action}", "{label}"{valuestr}])\',{delay});\n'
#    result =  eventString.format(category=str(category),
#                                 action=str(action),
#                                 label=str(label),
#                                 valuestr=valueString,
#                                 delay=totalDelay)
#    totalDelay+=500
#    return result


def getPQRBaseFileName(filename):
    root, ext = splitext(filename)
    if ext.lower() == ".pqr":
        return root
    return filename


def sortDictByValue(inputdict):
    """
    Sort a dictionary by its values

    Parameters
        inputdict:  The dictionary to sort (inputdict)
    Returns
        items: The dictionary sorted by value (list)
    """
    items = [(v, k) for k, v in inputdict.items()]
    items.sort()
    items.reverse()
    items = [k for v, k in items]
    return items


def shortestPath(graph, start, end, path=[]):
    """
    Uses recursion to find the shortest path from one node to
    another in an unweighted graph.  Adapted from
    http://www.python.org/doc/essays/graphs.html .

    Parameters:
        graph: A mapping of the graph to analyze, of the form
               {0: [1,2], 1:[3,4], ...} . Each key has a list
               of edges.
        start: The ID of the key to start the analysis from
        end:   The ID of the key to end the analysis
        path:  Optional argument used during the recursive step
               to keep the current path up to that point

    Returns:
        (variable): Returns a list of the shortest path (list)
                    Returns None if start and end are not
                    connected
    """

    path = path + [start]
    if start == end:
        return path
    if not graph.has_key(start):
        return None
    shortest = None
    for node in graph[start]:
        if node not in path:
            newpath = shortestPath(graph, node, end, path)
            if newpath:
                if not shortest or len(newpath) < len(shortest):
                    shortest = newpath
    return shortest


def analyzeConnectivity(map, key):
    """
    Analyze the connectivity of a given map using the key value.

    Parameters
        map:  The map to analyze (dict)
        key:  The key value (variable)
    Returns
        list: A list of connected values to the key (list)
    """
    list = []
    keys = [key]
    while len(keys) > 0:
        key = keys[0]
        if key not in list:
            list.append(key)
            # The following 4 lines are modified by Greg Cipriano as a bug fix
            if key in map:
                for value in map[key]:
                    if value not in list:
                        keys.append(value)

        keys.pop(keys.index(key))

    return list


def getAngle(coords1, coords2, coords3):
    """
    Get the angle between three coordinates

    Parameters
        coords1:  The first coordinate set (atom)
        coords2:  The second (vertex) coordinate set (atom)
        coords3:  The third coordinate set (atom)
    Returns
        angle:  The angle between the atoms (float)
    """
    angle = 0.0
    c1 = subtract(coords3, coords2)
    c2 = subtract(coords1, coords2)
    norm1 = normalize(c1)
    norm2 = normalize(c2)
    dotted = dot(norm1, norm2)
    if dotted > 1.0:  # If normalized, this is due to rounding error
        dotted = 1.0
    rad = abs(math.acos(dotted))
    angle = rad * 180.0 / math.pi
    if angle > 180.0:
        angle = 360.0 - angle
    return angle


# TODO: with changes to --userff and --usernames getFFfile and getNamesFile do not need to go on wild
# goose chases to find the files in question.
def getFFfile(name):
    """
    Grab the forcefield file.  May or may not residue in the dat/
    directory.
    """

    if name is None:
        return ""

    path = ""
    dirs = sys.path + ["dat"]
    if name in ["amber", "charmm", "parse", "tyl06", "peoepb", "swanson", "gromos"]:
        name = name.upper()

    names = ["dat/%s.DAT" % name]

    names.append("%s.DAT" % name)
    names.append("%s.dat" % name)
    names.append("dat/%s" % name)
    names.append(name)

    for guess in names:
        if os.path.isfile(guess):
            return guess

        for p in dirs:
            testpath = "%s/%s" % (p, guess)
            if os.path.isfile(testpath):
                return testpath

    # If we get here return empty string

    return ""


def getNamesFile(name):
    """
    Grab the *.names file that contains the XML mapping.

    Parameters
        name:  The name of the forcefield (string)
    Returns
        path:  The path to the file (string)
    """

    if name is None:
        return ""

    path = ""
    dirs = sys.path + ["dat"]
    if name in ["amber", "charmm", "parse", "tyl06", "peoepb", "swanson", "gromos"]:
        name = name.upper()

    names = ["dat/%s.names" % name]
    names.append("%s.names" % name)

    for guess in names:
        if os.path.isfile(guess):
            return guess

        for p in dirs:
            testpath = "%s/%s" % (p, guess)
            if os.path.isfile(testpath):
                return testpath

    # If we get here return empty string

    return ""


def getDatFile(name):
    """
    Grab a data file. If the file cannot be found in the
    given directory, try the current system path.

    Parameters
        name:  The name of the file to get (string)
    Returns
        path:  The path to the file (string)
    """
    path = ""

    if os.path.isfile(name):
        path = name

    for p in sys.path:
        testpath = "%s/%s" % (p, name)
        if os.path.isfile(testpath):
            path = testpath

    return path


def getPDBFile(path):
    """
    Obtain a PDB file.  First check the path given on the command
    line - if that file is not available, obtain the file from the
    PDB webserver at http://www.rcsb.org/pdb/ .

    Parameters
        path:  Name of PDB file to obtain (string)

    Returns
        file:  File object containing PDB file (file object)
    """

    import os, urllib

    file = None
    if not os.path.isfile(path):
        URLpath = "https://files.rcsb.org/download/" + path + ".pdb"
        try:
            file = urllib.urlopen(URLpath)
            if file.getcode() != 200 or "nosuchfile" in file.geturl():
                raise IOError
        except IOError:
            return None
    else:
        file = open(path, "rU")
    return file


def distance(coords1, coords2):
    """
    Calculate the distance between two coordinates, as denoted by

        dist = sqrt((x2- x1)^2 + (y2 - y1)^2 + (z2 - z1)^2))

    Parameters
        coords1: Coordinates of form [x,y,z]
        coords2: Coordinates of form [x,y,z]
    Returns
        dist:  Distance between the two coordinates (float)
    """
    dist = 0.0
    list = []

    p = coords2[0] - coords1[0]
    q = coords2[1] - coords1[1]
    r = coords2[2] - coords1[2]
    dist = math.sqrt(p * p + q * q + r * r)

    return dist


def add(coords1, coords2):
    """
    Add one 3-dimensional point to another

    Parameters
        coords1: coordinates of form [x,y,z]
        coords2: coordinates of form [x,y,z]
    Returns
        list:  List of coordinates equal to coords2 + coords1 (list)
    """
    x = coords1[0] + coords2[0]
    y = coords1[1] + coords2[1]
    z = coords1[2] + coords2[2]
    return [x, y, z]


def subtract(coords1, coords2):
    """
    Subtract one 3-dimensional point from another

    Parameters
        coords1: coordinates of form [x,y,z]
        coords2: coordinates of form [x,y,z]
    Returns
        list:  List of coordinates equal to coords1 - coords2 (list)
    """
    x = coords1[0] - coords2[0]
    y = coords1[1] - coords2[1]
    z = coords1[2] - coords2[2]
    return [x, y, z]


def cross(coords1, coords2):
    """
    Find the cross product of two 3-dimensional points

    Parameters
        coords1: coordinates of form [x,y,z]
        coords2: coordinates of form [x,y,z]
    Returns
        list:  Cross product coords2 and coords1 (list)
    """
    list = []
    x = coords1[1] * coords2[2] - coords1[2] * coords2[1]
    y = coords1[2] * coords2[0] - coords1[0] * coords2[2]
    z = coords1[0] * coords2[1] - coords1[1] * coords2[0]
    list = [x, y, z]
    return list


def dot(coords1, coords2):
    """
    Find the dot product of two 3-dimensional points

    Parameters
        coords1: coordinates of form [x,y,z]
        coords2: coordinates of form [x,y,z]
    Returns
        value:  Dot product coords2 and coords1 (float)
    """
    value = 0.0
    for i in range(3):
        value += coords1[i] * coords2[i]
    return value


def normalize(coords):
    """
    Normalize a set of coordinates

    Parameters
        coords: coordinates of form [x,y,z]
    Returns
        list: normalized coordinates (list)
    """
    list = []
    dist = math.sqrt(pow(coords[0], 2) + pow(coords[1], 2) + pow(coords[2], 2))
    if dist > SMALL:
        a = coords[0] / dist
        b = coords[1] / dist
        c = coords[2] / dist
        list = [a, b, c]
    else:
        list = coords
    return list


def factorial(n):
    """
    Returns the factorial of the given number n
    """
    if n <= 1:
        return 1
    return n * factorial(n - 1)


def getDihedral(coords1, coords2, coords3, coords4):
    """
    Calculate the angle using the four atoms

    Parameters
        coords1: First of four coordinates of form [x,y,z]
        coords2: Second of four
        coords3: Third of four
        coords4: Fourth of four
    Returns
        value: Size of the angle (float)
    """
    value = 0.0

    list43 = subtract(coords4, coords3)
    list32 = subtract(coords3, coords2)
    list12 = subtract(coords1, coords2)

    A = cross(list12, list32)
    Anorm = normalize(A)
    B = cross(list43, list32)
    Bnorm = normalize(B)

    scal = dot(Anorm, Bnorm)
    if abs(scal + 1.0) < SMALL:
        value = 180.0
    elif abs(scal - 1.0) < SMALL:
        value = 0.0
    else:
        value = DIHEDRAL * math.acos(scal)

    chiral = dot(cross(Anorm, Bnorm), list32)
    if chiral < 0:
        value = value * -1.0
    return value



"""
    Resinter extension

    Print interaction energy between each residue pair in the protein. 
"""

__date__ = "21 October 2011"
__authors__ = "Kyle Monson and Emile Hogan"

import extensions
from src.hydrogens import Optimize

# itertools FTW!
from itertools import product, permutations, izip, count
from collections import defaultdict
from src.hydrogens import hydrogenRoutines

_titrationSets = (
    ("ARG", "AR0"),
    ("ASP", "ASH"),
    ("CYS", "CYX"),
    ("GLU", "GLH"),
    ("HSD", "HSE", "HSP"),
    ("HID", "HIE", "HIP"),
    ("LYN", "LYS"),
    ("TYR", "TYM"),
    ("NEUTRAL-CTERM", "CTERM"),
    ("NEUTRAL-NTERM", "NTERM"),
)

_titrationSetsMap = defaultdict(tuple)

for tsSet in _titrationSets:
    for ts in tsSet:
        _titrationSetsMap[ts] = tsSet

# loose ends.
_titrationSetsMap["HIS"] = _titrationSetsMap["HSD"]
_titrationSetsMap["CYM"] = _titrationSetsMap["CYS"]

_pairEnergyResults = {}
# If the residue pair energy for a specific pair changes less than this ignore it.
PAIR_ENERGY_EPSILON = 1.0e-14


def addExtensionOptions(extensionGroup):
    """
    Add options.
    """
    extensionGroup.add_option(
        "--residue_combinations",
        dest="residue_combinations",
        action="store_true",
        default=False,
        help="""Remap residues to different titration states and rerun resinter appending output.
Consider only the minimum number of whole protein titration combinations needed to
test each possible pairing of residue titration states. Normally used with
--noopt. If a protein titration state combination results in a pair of residue being 
re-tested in the same individual titration states a warning will be generated if the 
re-tested result is different. This warning should not be possible if used with --noopt.""",
    )

    extensionGroup.add_option(
        "--all_residue_combinations",
        dest="all_residue_combinations",
        action="store_true",
        default=False,
        help="""Remap residues to ALL possible titration state combinations and rerun resinter appending output.
Results with --noopt should be the same as --residue_combinations. Runs considerably slower than
--residue_combinations and generates the same type of warnings. 
Use without --noopt to discover how hydrogen optimization affects residue 
interaction energies via the warnings in the output.""",
    )


def usage():
    """
    Returns usage text for resinter.
    """
    txt = "Print interaction energy between each residue pair in the protein to {output-path}.resinter."
    return txt


def _combinations(sublist, remainder):
    """
    combinations function helper

    sublist - first list in list of lists for combinations
    remainder - list of remaining lists
    """
    if remainder and sublist:
        for item in sublist:
            for result in _combinations(remainder[0], remainder[1:]):
                yield [item] + result
        return
    elif sublist:
        for item in sublist:
            yield [item]
        return
    elif remainder:
        for result in _combinations(remainder[0], remainder[1:]):
            yield [None] + result
        return

    yield [None]


def combinations(initialList):
    """
    Wrapper for main combinations function.

    Iterates over each possible combination of single items
    from each sub-list. For example:

    combinations([[1,2],[3,4]] -> [1,3], [1,4], [2,3], [2,4] in that order.


    initialList - list of lists to derive combinations from.
    """
    if not initialList:
        return
    for result in _combinations(initialList[0], initialList[1:]):
        yield result


def pairwiseCombinations(initialList):
    """
    Creates the minimum set of combinations that will make available
    every possible pair available.
    """
    r = [len(x) for x in initialList]
    m = min(r)
    M = max(r)
    n = len(initialList)

    R = set()

    for i in range(m):
        t = [initialList[x][i] for x in range(n)]
        R.add(tuple(t))

    for i in range(m, M):
        t = [initialList[x][min(r[x] - 1, i)] for x in range(n)]
        R.add(tuple(t))

    for i in range(m):
        for j in range(n):
            for k in range(i + 1, r[j]):
                prejth = [initialList[x][i] for x in range(j)]

                jth = initialList[j][k]

                postjth = [initialList[x][i] for x in range(j + 1, n)]

                t = prejth + [jth] + postjth
                R.add(tuple(t))

    for i in range(m, M):
        for j in range(n):
            if r[i] < i:
                continue
            for k in range(i + 1, r[j]):
                prejth = [initialList[x][min(r[x] - 1, i)] for x in range(j)]

                jth = initialList[j][k]

                postjth = [initialList[x][min(r[x] - 1, i)] for x in range(j + 1, n)]

                t = prejth + [jth] + postjth
                R.add(tuple(t))

    return R


def get_residue_interaction_energy(residue1, residue2):
    """
    Returns to total energy of every atom pair between the two residues.

    Uses Optimize.getPairEnergy and it's donor/accepter model
    to determine energy.

    residue1 - "donor" residue
    residue2 - "acceptor" residue

    THE RESULTS OF THIS FUNCTION ARE NOT SYMMETRIC. Swapping
    residue1 and residue2 will not always produce the same result.
    """
    energy = 0.0
    for pair in product(residue1.getAtoms(), residue2.getAtoms()):
        energy += Optimize.getPairEnergy(pair[0], pair[1])

    return energy


def save_residue_interaction_energies(residues, output):
    """
    Writes out the residue interaction energy for each possible
    residue pair in the protein.
    """
    residuepairs = permutations(residues, 2)

    for pair in residuepairs:
        energy = get_residue_interaction_energy(pair[0], pair[1])
        pairText = str(pair[0]) + " " + str(pair[1])
        if pairText in _pairEnergyResults:

            oldEnergy = _pairEnergyResults[pairText]
            energyDiff = oldEnergy - energy
            if abs(energyDiff) > PAIR_ENERGY_EPSILON:
                txt = "#%s re-tested" % pairText
                txt += " with a difference of %s" % repr(energyDiff)
                if energy != 0:
                    txt += " and a reference of %s" % repr(energyDiff / energy)
                else:
                    txt += " and the previous energy was 0"
                txt += "\n"

                output.write(txt)

            continue

        _pairEnergyResults[pairText] = energy


def get_residue_titration_sets(residues):
    """
    Returns all possible titration states for each residue as a list of lists.
    """
    result = []
    for residue in residues:
        result.append(_titrationSetsMap.get(residue.name, (residue.name,)))

    return result


def process_residue_set(
    residueSet,
    routines,
    output,
    clean=False,
    neutraln=False,
    neutralc=False,
    ligand=None,
    assign_only=False,
    chain=False,
    debump=True,
    opt=True,
):
    routines.write(str(residueSet) + "\n")

    routines.removeHydrogens()

    for newResidueName, oldResidue, index in izip(
        residueSet, routines.protein.getResidues(), count()
    ):
        if newResidueName is None:
            continue

        chain = routines.protein.chainmap[oldResidue.chainID]
        chainIndex = chain.residues.index(oldResidue)
        residueAtoms = oldResidue.atoms

        # Create the replacement residue
        newResidue = routines.protein.createResidue(residueAtoms, newResidueName)

        # Make sure our names are cleaned up for output.
        newResidue.renameResidue(newResidueName)

        # Drop it in
        routines.protein.residues[index] = newResidue
        chain.residues[chainIndex] = newResidue

    # Run the meaty bits of PDB2PQR
    routines.setTermini(neutraln, neutralc)
    routines.updateBonds()

    if not clean and not assign_only:
        routines.updateSSbridges()

        if debump:
            routines.debumpProtein()

        routines.addHydrogens()

        hydRoutines = hydrogenRoutines(routines)

        if debump:
            routines.debumpProtein()

        if opt:
            hydRoutines.setOptimizeableHydrogens()
            hydRoutines.initializeFullOptimization()
            hydRoutines.optimizeHydrogens()
        else:
            hydRoutines.initializeWaterOptimization()
            hydRoutines.optimizeHydrogens()

        # Special for GLH/ASH, since both conformations were added
        hydRoutines.cleanup()

    save_residue_interaction_energies(routines.protein.getResidues(), output)


def write_all_residue_interaction_energies_combinations(
    routines, output, options, all_residue_combinations=False
):
    """
    For every titration state combination of residue output the
    interaction energy for all possible residue pairs.
    """
    residueNamesList = get_residue_titration_sets(routines.protein.getResidues())

    routines.write("Testing the following combinations\n")
    namelist = [r.name for r in routines.protein.getResidues()]
    combinationsData = izip(namelist, residueNamesList)
    for thing in combinationsData:
        routines.write(str(thing) + "\n")

    if all_residue_combinations:
        combinationGenerator = combinations(residueNamesList)
    else:
        combinationGenerator = pairwiseCombinations(residueNamesList)

    count = 0
    for residueSet in combinationGenerator:
        count += 1
        process_residue_set(
            residueSet,
            routines,
            output,
            clean=options.clean,
            neutraln=options.neutraln,
            neutralc=options.neutralc,
            ligand=options.ligand,
            assign_only=options.assign_only,
            chain=options.chain,
            debump=options.debump,
            opt=options.opt,
        )

    for resultKey in sorted(_pairEnergyResults.iterkeys()):
        output.write(resultKey + " " + str(_pairEnergyResults[resultKey]) + "\n")

    routines.write(str(count) + " residue combinations tried\n")


def create_resinter_output(
    routines,
    outfile,
    options,
    residue_combinations=False,
    all_residue_combinations=False,
):
    """
    Output the interaction energy between each possible residue pair.
    """
    routines.write("Printing residue interaction energies...\n")

    output = extensions.extOutputHelper(routines, outfile)

    if residue_combinations or all_residue_combinations:
        write_all_residue_interaction_energies_combinations(
            routines, output, options, all_residue_combinations=all_residue_combinations
        )
    else:
        save_residue_interaction_energies(routines.protein.getResidues(), output)


def run_extension(routines, outroot, options):
    outname = outroot + ".resinter"
    with open(outname, "w") as outfile:
        create_resinter_output(
            routines,
            outfile,
            options,
            residue_combinations=options.residue_combinations,
            all_residue_combinations=options.all_residue_combinations,
        )
